/* anemon, Copyright (c) 2001 Gabriel Finch
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 06/02
  */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

int resized;


#define TWO_PI     (2.0 * M_PI)
#define RND(x)     (rand() % (x))
#define MAXPEND    2000
#define MAXPTS    200
#define TRUE 1
#define FALSE 0




static int
arms,                       /* number of arms */
finpoints;                  /* final number of points in each array. */
static long delay;              /* usecs to wait between updates. */

static int         scrWidth, scrHeight;

typedef struct {
	double x,y,z;
	int sx,sy,sz;
} vPend;

typedef unsigned short bool;

static bool dbuf;
static int width;

typedef struct {
	Image * col;
	int numpt;
	int growth;
	unsigned short rate;
} appDef;

static vPend *vPendage;  /* 3D representation of appendages */
static appDef *appD;  /* defaults */
static vPend *vCurr, *vNext;
static appDef *aCurr;

static double turn, turndelta;

static int    mx, my;            /* max screen coordinates. */
static int withdraw;

static void *
xmalloc(uint size)
{
	void *ret;

	if ((ret = malloc(size)) == NULL) {
		sysfatal("anemone: out of memory %r\n");
	}
	return ret;
}


static void
initAppendages(void)
{
	int    i;
	/*int    marginx, marginy; */

	/*double scalex, scaley;*/

	double x,y,z,dist;

	mx = scrWidth - 1;
	my = scrHeight - 1;

	/* each appendage will have: colour,
     number of points, and a grow or shrink indicator */

	/* added: growth rate 1-10 (smaller==faster growth) */
	/* each appendage needs virtual coords (x,y,z) with y and z combining to
     give the screen y */

	vPendage = (vPend *) xmalloc((finpoints + 1) * sizeof(vPend) * arms);
	appD = (appDef *) xmalloc(sizeof(appDef) * arms);


	for (i = 0; i < arms; i++) {
		aCurr = appD + i;
		vCurr = vPendage + (finpoints + 1) * i;
		vNext = vCurr + 1;

		aCurr->col = colors[(long)RND(256)];
		aCurr->numpt = 1;
		aCurr->growth=finpoints/2+RND(finpoints/2);
		aCurr->rate=RND(11)*RND(11);

		dist=1.;

		do 
		{
			x=(1-RND(1001)/500);
			y=(1-RND(1001)/500);
			z=(1-RND(1001)/500);
			dist=x*x+y*y+z*z;
		}		while (dist>=1.);

		vCurr->x=x*200;
		vCurr->y=my/2+y*200;
		vCurr->z=0+z*200;

		/* start the arm going outwards */
		vCurr->sx=vCurr->x/5;
		vCurr->sy=(vCurr->y-my/2)/5;
		vCurr->sz=(vCurr->z)/5;


		vNext->x=vCurr->x+vCurr->sx;
		vNext->y=vCurr->y+vCurr->sy;
		vNext->z=vCurr->z+vCurr->sz;
	}
}

static void
initAnemone(void)
{

	turn = 0.0;

	width = 2;
	arms = 128;
	finpoints = 64;
	delay = 40000;
	withdraw = 1200;
	turndelta = 50.0/100000;

	dbuf=TRUE;




	scrWidth = Dx(screen->r);
	scrHeight = Dy(screen->r);

	initAppendages();
}


static void
createPoints(void)
{
	int i;
	int withdrawall=RND(withdraw);

	for (i = 0; i< arms; i++) {
		aCurr = appD + i;
		if (!withdrawall) {
			aCurr->growth=-finpoints;
			turndelta=-turndelta;
		} else if (withdrawall<11)
			aCurr->growth=-aCurr->numpt;
		else if (RND(100)<aCurr->rate) {
			if (aCurr->growth>0) {
				if (!(--aCurr->growth)) aCurr->growth=-RND(finpoints)-1;
				vCurr = vPendage + (finpoints + 1) * i + aCurr->numpt-1;
				if (aCurr->numpt<finpoints - 1) {
					/* add a piece */
					vNext=vCurr + 1;
					aCurr->numpt++;
					vNext->sx=vCurr->sx+RND(3)-1;
					vNext->sy=vCurr->sy+RND(3)-1;
					vNext->sz=vCurr->sz+RND(3)-1;
					vCurr=vNext+1;
					vCurr->x=vNext->x+vNext->sx;
					vCurr->y=vNext->y+vNext->sy;
					vCurr->z=vNext->z+vNext->sz;
				}
			}
		}
	}
}


static void
drawImage(double sint, double cost)
{
	int q,numpt;
	double cx,cy,cz,nx,ny,nz;
	Point p1, p2;

	if ((numpt=aCurr->numpt)==1) return;
	//  XSetForeground(dpy, gcDraw, aCurr->col);

	vNext=vCurr+1;

	cx=vCurr->x;
	cy=vCurr->y;
	cz=vCurr->z;


	for (q = 0; q < numpt-1; q++) {
		nx=vNext->x+2-RND(5);
		ny=vNext->y+2-RND(5);
		nz=vNext->z+2-RND(5);

		p1 = Pt(mx/2+
			cx*cost-cz*sint, 
			cy);
		p2 = Pt( 
			mx/2+
			nx*cost-
			nz*sint, 
			ny);
		line(img, p1, p2, Endsquare, Endsquare, 1, aCurr->col, ZP);
		vCurr++;
		vNext++;

		cx=nx;
		cy=ny;
		cz=nz;
	}

	p1 = Pt(mx/2+cx*cost-cz*sint, cy);
	p2 = Pt(mx/2+nx*cost-nz*sint, ny);
	//line(screen, p1, p2, Endsquare, Endsquare, 3, aCurr->col, ZP);
	fillellipse(img, p1, 
		3, 3, aCurr->col, ZP);



}

static void
animateAnemone(void)
{
	int i;
	double sint=sin(turn),cost=cos(turn);

	aCurr = appD;
	for (i = 0; i< arms; i++) {
		vCurr=vPendage + (finpoints + 1) * i;
		if (RND(25)<aCurr->rate) {
			if (aCurr->growth<0) {
				aCurr->numpt-=aCurr->numpt>1;
				if (!(++aCurr->growth)) aCurr->growth=RND(finpoints-aCurr->numpt)+1;
			}
		}
		drawImage(sint, cost);
		turn+=turndelta;
		aCurr++;
	}
	createPoints();
	//  sleep(delay);
	draw(screen, screen->r, img, nil, ZP);
	frames++;
	flushimage(display, 1);

	if (turn>=TWO_PI) turn-=TWO_PI;
}

void
screenhack(void)
{
	Mouse m;


	for (;;) {
		if(resized) {
			resized = 0;
			initAnemone();
		}

		draw(img, img->r, display->black, nil, ZP);
		animateAnemone();
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(nil);
			}
		}
	}
}



void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	smooth(colors);

	resized = 1;

}


void 
threadmain(int argc, char **argv)
{
	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;

	srand(time(0));


	if(initdraw(nil, nil, "anemone") < 0)
		sysfatal("initdraw failed: %r");


	einit(Emouse);
	eresized(0);
	screenhack();
}












