/* coral, by "Frederick G.M. Roeber" <roeber@netscape.com>, 15-jul-97.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
  * Ported to Plan 9 by andrey@lanl.gov, 07/02
  */


/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	fprint(2, "sorry, cannot resize\n");
	threadexitsall("resize");
}


static unsigned int default_fg_pixel;
#define NCOLORSMAX 256
static int ncolors = 0;
static int colorindex = 0;
static int colorsloth;

static XPoint *walkers = 0;
static int nwalkers;
static int width, widthb;
static int height;

static unsigned int *board = 0;
#define getdot(x,y) (board[(y*widthb)+(x>>5)] &  (1<<(x & 31)))
#define setdot(x,y) (board[(y*widthb)+(x>>5)] |= (1<<(x & 31)))


static void
init_coral(void)
{
	int seeds;
	int density;
	int i;

	width = Dx(screen->r);
	widthb = ((Dx(screen->r) + 31) >> 5);
	height = Dy(screen->r);
	if (board) free(board);
	board = (unsigned int *)calloc(widthb * Dy(screen->r), sizeof(unsigned int));
	if(!board) threadexitsall("!board");


	ncolors = NCOLORSMAX;

	colorindex = random()%ncolors;

	density = 25;
	if( density < 1 ) density = 1;
	if( density > 100 ) density = 90; /* more like mold than coral */
	nwalkers = (width*height*density)/100;
	if (walkers) free(walkers);
	walkers = (XPoint *)calloc(nwalkers, sizeof(XPoint));
	if( (XPoint *)0 == walkers ) threadexitsall("walkers == nil");

	seeds = 20;
	if( seeds < 1 ) seeds = 1;
	if( seeds > 1000 ) seeds = 1000;

	colorsloth = nwalkers*2/ncolors;

	for( i = 0; i < seeds; i++ ) {
		int x, y;
		Rectangle r;
		do 
		{
			x = 1 + random() % (width - 2);
			y = 1 + random() % (height - 2);
		}		while( getdot(x, y) );

		setdot((x-1), (y-1));
		setdot(x, (y-1));
		setdot((x+1), (y-1));
		setdot((x-1),  y   );
		setdot(x,  y   );
		setdot((x+1),  y   );
		setdot((x-1), (y+1));
		setdot(x, (y+1));
		setdot((x+1), (y+1));
		//        XDrawPoint(dpy, window, draw_gc, x, y);

		r = Rpt(Pt(screen->r.min.x + x, screen->r.min.y + y), 
			Pt(screen->r.min.x + x+1, screen->r.min.y + y+1))	;

		draw(screen, r, colors[colorindex], nil, ZP);

	}

	for( i = 0; i < nwalkers; i++ ) {
		walkers[i].x = (random() % (width-2)) + 1;
		walkers[i].y = (random() % (height-2)) + 1;
	}
}


/* returns 2 bits of randomness (conserving calls to random()).
   This speeds things up a little, but not a lot (5-10% or so.)
 */
static int 
rand_2(void)
{
	static int i = 0;
	static int r = 0;
	if (i != 0) {
		i--;
	}
	else {
		i = 15;
		r = random();
	}

	{
		register int j = (r & 3);
		r = r >> 2;
		return j;
	}
}


static void
coral(void)
{
	int delay2 = 10;

	int max_points = 200;
	int npoints = 0;
	XPoint *pointbuf = (XPoint *) calloc(sizeof(XPoint), max_points+2);
	if (!pointbuf) threadexitsall("!pointbuf");

	while( 1 ) {
		int i,i2;

		for( i = 0; i < nwalkers; i++ ) {
			int x = walkers[i].x;
			int y = walkers[i].y;

			if( getdot(x, y) ) {

				Bool flush = False;
				Bool color = False;

				/* XDrawPoint(dpy, window, draw_gc, x, y); */
				pointbuf[npoints].x = x;
				pointbuf[npoints].y = y;
				npoints++;

				/* Mark the surrounding area as "sticky" */
				setdot((x-1), (y-1));
				setdot(x, (y-1));
				setdot((x+1), (y-1));
				setdot((x-1),  y   );
				setdot((x+1),  y   );
				setdot((x-1), (y+1));
				setdot(x, (y+1));
				setdot((x+1), (y+1));
				nwalkers--;
				walkers[i].x = walkers[nwalkers].x;
				walkers[i].y = walkers[nwalkers].y;
				if( 0 == (nwalkers%colorsloth) ) {
					color = True;
				}

				if (flush || color || 0 == nwalkers || npoints >= max_points) {
					//		  XDrawPoints(dpy, window, draw_gc, pointbuf, npoints,
					//			      CoordModeOrigin);
					for(i2 = 0; i2 < npoints; i2++) {
						Rectangle r;
						Point p = Pt(screen->r.min.x, screen->r.min.y);

						p = addpt(p, pointbuf[i2]);
						r = Rpt(p, addpt(p, Pt(1,1)));
						draw(screen, r, colors[colorindex], nil, ZP);

					}

					npoints = 0;
				}

				if (color) {
					colorindex++;
					if( colorindex == ncolors )
						colorindex = 0;
				}

				if( 0 == nwalkers ) {
					free(pointbuf);
					return;
				}
			}
			else {
				/* move it a notch */
				do 
				{
					switch(rand_2()) {
					case 0:
						if( 1 == x ) continue;
						walkers[i].x--;
						break;
					case 1:
						if( width-2 == x ) continue;
						walkers[i].x++;
						break;
					case 2:
						if( 1 == y ) continue;
						walkers[i].y--;
						break;
					default: /* case 3: */
						if( height-2 == y ) continue;
						walkers[i].y++;
						break;
						/* default:
		      abort(); */
					}
				}	while(0);
			}
		}

		if (delay2 > 0) {
			if (npoints > 0) {
				//	    XDrawPoints(dpy, window, draw_gc, pointbuf, npoints,
				//			CoordModeOrigin);
				for(i2 = 0; i2 < npoints; i2++) {
					Rectangle r;
					Point p = Pt(screen->r.min.x, screen->r.min.y);

					p = addpt(p, pointbuf[i2]);
					r = Rpt(p, addpt(p, Pt(1,1)));
					draw(screen, r, colors[colorindex], nil, ZP);

				}
				npoints = 0;
			}
			if(ecanmouse()) {
				m = emouse();
				if(m.buttons&4)
					if(emenuhit(3, &m, &menu) == 0) {
						threadexitsall(0);
					}
			}
			frames++;
			flushimage(display, 1);
			sleep(delay2);
		}
	}
}


void
screenhack(void)
{
	int delay = 5000;
	while( 1 ) {
		init_coral();
		coral();
		if( delay ) sleep(delay);

		draw(screen, screen->r, display->black, nil , ZP);

	}
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "coral") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);

	einit(Emouse);

	draw(screen, screen->r, display->black, nil, ZP);

	screenhack();
}
