/* xscreensaver, Copyright (c) 1999, 2001, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"

#define countof(x) (sizeof(x)/sizeof(*(x)))

static Bool transparent_p;
static int nplanes;
static unsigned long base_pixel, *plane_masks;

struct throbber {
	int x, y;
	int size;
	int max_size;
	int thickness;
	int speed;
	int fuse;
	long gc;
	void (*draw) (struct throbber *);
};

static void
draw_star (struct throbber *t)
{
	XPoint points[11];
	int x = t->x;
	int y = t->y;
	int s = t->size / 0.383;  /* trial and error, I forget how to derive this */
	int s2 = t->size;
	double c = M_PI * 2;
	double o = -M_PI / 2;

	points[0].x = x + s  * cos(o + 0.0*c);
	points[0].y = y + s  * sin(o + 0.0*c);
	points[1].x = x + s2 * cos(o + 0.1*c);
	points[1].y = y + s2 * sin(o + 0.1*c);
	points[2].x = x + s  * cos(o + 0.2*c);
	points[2].y = y + s  * sin(o + 0.2*c);
	points[3].x = x + s2 * cos(o + 0.3*c);
	points[3].y = y + s2 * sin(o + 0.3*c);
	points[4].x = x + s  * cos(o + 0.4*c);
	points[4].y = y + s  * sin(o + 0.4*c);
	points[5].x = x + s2 * cos(o + 0.5*c);
	points[5].y = y + s2 * sin(o + 0.5*c);
	points[6].x = x + s  * cos(o + 0.6*c);
	points[6].y = y + s  * sin(o + 0.6*c);
	points[7].x = x + s2 * cos(o + 0.7*c);
	points[7].y = y + s2 * sin(o + 0.7*c);
	points[8].x = x + s  * cos(o + 0.8*c);
	points[8].y = y + s  * sin(o + 0.8*c);
	points[9].x = x + s2 * cos(o + 0.9*c);
	points[9].y = y + s2 * sin(o + 0.9*c);
	points[10] = points[0];

	//XDrawLines (dpy, w, t->gc, points, countof(points), CoordModeOrigin);
	poly(img, points, countof(points), Enddisc, Enddisc, t->thickness, colors[t->gc], ZP);
}

static void
draw_circle (struct throbber *t)
{
	//XDrawArc (dpy, w, t->gc,
	//   t->x - t->size / 2,
	//    t->y - t->size / 2,
	//    t->size, t->size,
	//    0, 360*64);
	ellipse(img, Pt(t->x ,t->y), 
		t->size, t->size, t->thickness, colors[t->gc], ZP);
}

static void
draw_hlines (struct throbber *t)
{
	//XDrawLine (dpy, w, t->gc, 0,
	//    t->y - t->size, t->max_size,t->y - t->size);
	line(img, Pt(0,t->y - t->size), 
		Pt(t->max_size,t->y - t->size), 
		Enddisc, Enddisc, t->thickness, colors[t->gc], ZP);
	//XDrawLine (dpy, w, t->gc, 0,
	//    t->y + t->size, t->max_size,
	//    t->y + t->size);
	line(img, Pt(0,t->y + t->size), 
		Pt(t->max_size,t->y + t->size), 
		Enddisc, Enddisc, t->thickness, colors[t->gc], ZP);
}

static void
draw_vlines (struct throbber *t)
{
	//XDrawLine (dpy, w, t->gc,
	//    t->x - t->size, 0,
	//    t->x - t->size, t->max_size);
	line(img, Pt(t->x - t->size, 0), 
		Pt(t->x - t->size, t->max_size), 
		Enddisc, Enddisc, t->thickness, colors[t->gc], ZP);
	//XDrawLine (dpy, w, t->gc,
	//    t->x + t->size, 0,
	//    t->x + t->size, t->max_size);
	line(img, Pt(t->x + t->size, 0), 
		Pt(t->x + t->size, t->max_size), 
		Enddisc, Enddisc, t->thickness, colors[t->gc], ZP);
}

static void
draw_corners (struct throbber *t)
{
	int s = (t->size + t->thickness) / 2;
	XPoint points[3];

	points[0].x = 0;
	points[0].y = t->y - s;
	points[1].x = t->x - s;
	points[1].y = t->y - s;
	points[2].x = t->x - s;
	points[2].y = 0;
	//XDrawLines (dpy, w, t->gc, points, countof(points), CoordModeOrigin);
	poly(img, points, countof(points), Enddisc, Enddisc, t->thickness, colors[t->gc], screen->r.min);

	points[0].x = 0;
	points[0].y = t->y + s;
	points[1].x = t->x - s;
	points[1].y = t->y + s;
	points[2].x = t->x - s;
	points[2].y = t->max_size;
	//XDrawLines (dpy, w, t->gc, points, countof(points), CoordModeOrigin);
	poly(img, points, countof(points), Enddisc, Enddisc, t->thickness, colors[t->gc], screen->r.min);

	points[0].x = t->x + s;
	points[0].y = 0;
	points[1].x = t->x + s;
	points[1].y = t->y - s;
	points[2].x = t->max_size;
	points[2].y = t->y - s;
	//XDrawLines (dpy, w, t->gc, points, countof(points), CoordModeOrigin);
	poly(img, points, countof(points), Enddisc, Enddisc, t->thickness, colors[t->gc], screen->r.min);

	points[0].x = t->x + s;
	points[0].y = t->max_size;
	points[1].x = t->x + s;
	points[1].y = t->y + s;
	points[2].x = t->max_size;
	points[2].y = t->y + s;
	//XDrawLines (dpy, w, t->gc, points, countof(points), CoordModeOrigin);
	poly(img, points, countof(points), Enddisc, Enddisc, t->thickness, colors[t->gc], screen->r.min);
}



static struct throbber *
make_throbber (int w, int h)
{
	struct throbber *t = (struct throbber *) malloc (sizeof (*t));
	t->x = w / 2;
	t->y = h / 2;
	t->max_size = w;
	t->speed = 15;
	t->fuse = 1 + (random() % 4);
	t->thickness = nrand(30)+15;

	if (t->speed < 0) t->speed = -t->speed;
	t->speed += (((random() % t->speed) / 2) - (t->speed / 2));
	if (t->speed > 0) t->speed = -t->speed;

	if (random() % 4)
		t->size = t->max_size;
	else
		t->size = t->thickness, t->speed = -t->speed;

	t->gc = nrand(256);

	switch (random() % 11) {
	case 0:
	case 1:
	case 2:
	case 3:
		t->draw = draw_star;
		break;
	case 4:
	case 5:
	case 6:
	case 7:
		t->draw = draw_circle;
		break;
	case 8:
		t->draw = draw_hlines;
		break;
	case 9:
		t->draw = draw_vlines;
		break;
	case 10:
		t->draw = draw_corners;
		break;
	default:
		abort();
		break;
	}

	return t;
}

static int
throb (struct throbber *t)
{
	t->size += t->speed;
	if (t->size <= (t->thickness / 2))
	{
		t->speed = -t->speed;
		t->size += (t->speed * 2);
	}
	else if (t->size > t->max_size)
	{
		t->speed = -t->speed;
		t->size += (t->speed * 2);
		t->fuse--;
	}

	if (t->fuse <= 0)
	{
		//		XFreeGC (dpy, t->gc);
		memset (t, 0, sizeof(*t));
		free (t);
		return -1;
	}
	else
	{
		t->draw (t);
		return 0;
	}
}



void
screenhack (void)
{
	int count = 5;
	int delay = 50;
	int i;
	struct throbber **throbbers;
	transparent_p = 0;


	throbbers = (struct throbber **) calloc (count, sizeof(struct throbber *));
	for (i = 0; i < count; i++)
		throbbers[i] = make_throbber (Dx(screen->r), Dy(screen->r));

	while (1)
	{
		//XFillRectangle (dpy, b, erase_gc, 0, 0, xgwa.width, xgwa.height);
		draw(img, img->r, display->black, nil, ZP);

		for (i = 0; i < count; i++)
			if (throb (throbbers[i]) < 0)
				throbbers[i] = make_throbber (Dx(screen->r), Dy(screen->r));

		draw(screen, screen->r, img, nil, ZP);
		frames++;
		flushimage(display, 1);
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if (delay)
			sleep (delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	//fprint(2, "resize not supported...\n");
	//threadexitsall("resize not supported");

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

}

void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;

	srand(time(0));


	if(initdraw(nil, nil, "deluxe") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			cmap2rgba(i<<2));
	}

	einit(Emouse);
	eresized(0);
	screenhack();
}
