/* drift --- drifting recursive fractal cosmic flames */


/*-
 * Copyright (c) 1991 by Patrick J. Naughton.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 10-May-1997: Jamie Zawinski <jwz@jwz.org> compatible with xscreensaver
 * 01-Jan-1997: Moved new flame to drift.  Compile time options now run time.
 * 01-Jun-1995: Updated by Scott Draves.
 * 27-Jun-1991: vary number of functions used.
 * 24-Jun-1991: fixed portability problem with integer mod (%).
 * 06-Jun-1991: Written, received from Scott Draves <spot@cs.cmu.edu>
 */


/* 
  * ported to Plan 9 by andrey@lanl.gov, 07/02
  */

/* 
  * TODO: fix rainbow coloring
  */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

#define COUNT 1


#define DEF_GROW "False"
#define DEF_LISS "False"
static Bool grow;
static Bool liss;



#define MAXBATCH1	200	/* mono */
#define MAXBATCH2	20	/* color */
#define FUSE		10	/* discard this many initial iterations */
#define NMAJORVARS	7
#define MAXLEV 10

typedef struct {
	/* shape of current flame */
	int         nxforms;
	double      f[2][3][MAXLEV];	/* a bunch of non-homogeneous xforms */
	int         variation[10];	/* for each xform */

	/* Animation */
	double      df[2][3][MAXLEV];

	/* high-level control */
	int         mode;	/* 0->slow/single 1->fast/many */
	int         nfractals;	/* draw this many fractals */
	int         major_variation;
	int         fractal_len;	/* pts/fractal */
	int         color;
	int         rainbow;	/* more than one color per fractal
						   1-> computed by adding dimension to fractal */

	int         width, height;	/* of window */
	int         timer;

	/* draw info about current flame */
	int         fuse;	/* iterate this many before drawing */
	int         total_points;	/* draw this many pts before fractal ends */
	int         npoints;	/* how many we've computed but not drawn */
	XPoint      pts[MAXBATCH1];	/* here they are */
	unsigned long pixcol;
	/* when drawing in color, we have a buffer per color */
	int        *ncpoints;
	XPoint     *cpts;

	double      x, y, c;
	int         liss_time;
	Bool        grow, liss;

	short       lasthalf;
	long        saved_random_bits;
	int         nbits;
} driftstruct;

static driftstruct *drifts = (driftstruct *) NULL;

static short
halfrandom(driftstruct * dp, int mv)
{
	unsigned long r;

	if (dp->lasthalf) {
		r = dp->lasthalf;
		dp->lasthalf = 0;
	}
	else {
		r = LRAND();
		dp->lasthalf = (short) (r >> 16);
	}
	r = r % mv;
	return r;
}

static int
frandom(driftstruct * dp, int n)
{
	int         result;

	if (3 > dp->nbits) {
		dp->saved_random_bits = LRAND();
		dp->nbits = 31;
	}
	switch (n) {
	case 2:
		result = (int) (dp->saved_random_bits & 1);
		dp->saved_random_bits >>= 1;
		dp->nbits -= 1;
		return result;

	case 3:
		result = (int) (dp->saved_random_bits & 3);
		dp->saved_random_bits >>= 2;
		dp->nbits -= 2;
		if (3 == result)
			return frandom(dp, 3);
		return result;

	case 4:
		result = (int) (dp->saved_random_bits & 3);
		dp->saved_random_bits >>= 2;
		dp->nbits -= 2;
		return result;

	case 5:
		result = (int) (dp->saved_random_bits & 7);
		dp->saved_random_bits >>= 3;
		dp->nbits -= 3;
		if (4 < result)
			return frandom(dp, 5);
		return result;
	default:
		fprint(2, "bad arg to frandom\n");
	}
	return 0;
}

#define DISTRIB_A (halfrandom(dp, 7000) + 9000)
#define DISTRIB_B ((frandom(dp, 3) + 1) * (frandom(dp, 3) + 1) * 120000)
#define LEN(x) (sizeof(x)/sizeof((x)[0]))

static void
initmode(int mode)
{
	driftstruct *dp = &drifts[0];

#define VARIATION_LEN 14

	dp->mode = mode;

	dp->major_variation = halfrandom(dp, VARIATION_LEN);
	/*  0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 6 */
	dp->major_variation = ((dp->major_variation >= VARIATION_LEN >> 1) &&
		(dp->major_variation < VARIATION_LEN - 1)) ?
		(dp->major_variation + 1) >> 1 : dp->major_variation >> 1;

	if (dp->grow) {
		dp->rainbow = 0;
		if (mode) {
			if (!dp->color || halfrandom(dp, 8)) {
				dp->nfractals = halfrandom(dp, 30) + 5;
				dp->fractal_len = DISTRIB_A;
			}
			else {
				dp->nfractals = halfrandom(dp, 5) + 5;
				dp->fractal_len = DISTRIB_B;
			}
		}
		else {
			dp->rainbow = dp->color;
			dp->nfractals = 1;
			dp->fractal_len = DISTRIB_B;
		}
	}
	else {
		dp->nfractals = 1;
		dp->rainbow = dp->color;
		dp->fractal_len = 2000000;
	}
	dp->fractal_len = (dp->fractal_len * COUNT) / 20;

	draw(screen, screen->r, display->black, nil, ZP);
}

static void
pick_df_coefs(void)
{
	driftstruct *dp = &drifts[0];
	int         i, j, k;
	double      r;

	for (i = 0; i < dp->nxforms; i++) {

		r = 1e-6;
		for (j = 0; j < 2; j++)
			for (k = 0; k < 3; k++) {
				dp->df[j][k][i] = ((double) halfrandom(dp, 1000) / 500.0 - 1.0);
				r += dp->df[j][k][i] * dp->df[j][k][i];
			}
		r = (3 + halfrandom(dp, 5)) * 0.01 / sqrt(r);
		for (j = 0; j < 2; j++)
			for (k = 0; k < 3; k++)
				dp->df[j][k][i] *= r;
	}
}

static void
free_drift(driftstruct *dp)
{
	if (dp->ncpoints != NULL) {
		free((void *) dp->ncpoints);
		dp->ncpoints = (int *) NULL;
	}
	if (dp->cpts != NULL) {
		free((void *) dp->cpts);
		dp->cpts = (XPoint *) NULL;
	}
}

static void
initfractal(void)
{
	driftstruct *dp = &drifts[0];
	int         i, j, k;

#define XFORM_LEN 9

	dp->fuse = FUSE;
	dp->total_points = 0;

	if (!dp->ncpoints) {
		if ((dp->ncpoints = (int *) malloc(sizeof (int) * 256)) ==
			NULL) {
			free_drift(dp);
			return;
		}
	}
	if (!dp->cpts) {
		if ((dp->cpts = (XPoint *) malloc(MAXBATCH2 * sizeof (XPoint) *
			256)) == NULL) {
			free_drift(dp);
			return;
		}
	}


	dp->npoints = 0;
	dp->nxforms = halfrandom(dp, XFORM_LEN);
	/* 2, 2, 2, 3, 3, 3, 4, 4, 5 */
	dp->nxforms = (dp->nxforms >= XFORM_LEN - 1) + dp->nxforms / 3 + 2;

	dp->c = dp->x = dp->y = 0.0;
	if (dp->liss && !halfrandom(dp, 10)) {
		dp->liss_time = 0;
	}
	if (!dp->grow)
		pick_df_coefs();
	for (i = 0; i < dp->nxforms; i++) {
		if (NMAJORVARS == dp->major_variation)
			dp->variation[i] = halfrandom(dp, NMAJORVARS);
		else
			dp->variation[i] = dp->major_variation;
		for (j = 0; j < 2; j++)
			for (k = 0; k < 3; k++) {
				if (dp->liss)
					dp->f[j][k][i] = sin(dp->liss_time * dp->df[j][k][i]);
				else
					dp->f[j][k][i] = ((double) halfrandom(dp, 1000) / 500.0 - 1.0);
			}
	}

	dp->pixcol = halfrandom(dp, 256);

}


void
init_drift(void)
{
	driftstruct *dp;

	if (drifts == NULL) {
		if ((drifts = (driftstruct *) calloc(1,
			sizeof (driftstruct))) == NULL)
			return;
	}
	dp = &drifts[0];

	dp->width = Dx(screen->r);
	dp->height = Dy(screen->r);
	dp->color = 1;

	if (NRAND(3) == 0) {
		if (NRAND(3) == 0)
			dp->grow = True;
		else {
			dp->grow = False;
			dp->liss = (Bool) (LRAND() & 1);
		}
	}
	else {
		dp->grow = grow;
		if (dp->grow)
			dp->liss = False;
		else
			dp->liss = liss;
	}
	initmode(1);
	initfractal();
}

static void
iter(driftstruct * dp)
{
	int         i = frandom(dp, dp->nxforms);
	double      nx, ny, nc;


	if (i)
		nc = (dp->c + 1.0) / 2.0;
	else
		nc = dp->c / 2.0;

	nx = dp->f[0][0][i] * dp->x + dp->f[0][1][i] * dp->y + dp->f[0][2][i];
	ny = dp->f[1][0][i] * dp->x + dp->f[1][1][i] * dp->y + dp->f[1][2][i];


	switch (dp->variation[i]) {
	case 1:
		/* sinusoidal */
		nx = sin(nx);
		ny = sin(ny);
		break;
	case 2:
		{
			/* complex */
			double      r2 = nx * nx + ny * ny + 1e-6;

			nx = nx / r2;
			ny = ny / r2;
			break;
		}
	case 3:
		/* bent */
		if (nx < 0.0)
			nx = nx * 2.0;
		if (ny < 0.0)
			ny = ny / 2.0;
		break;
	case 4:
		{
			/* swirl */

			double      r = (nx * nx + ny * ny);	/* times k here is fun */
			double      c1 = sin(r);
			double      c2 = cos(r);
			double      t = nx;

			if (nx > 1e4 || nx < -1e4 || ny > 1e4 || ny < -1e4)
				ny = 1e4;
			else
				ny = c2 * t + c1 * ny;
			nx = c1 * nx - c2 * ny;
			break;
		}
	case 5:
		{
			/* horseshoe */
			double      r, c1, c2, t;

			/* Avoid atan2: DOMAIN error message */
			if (nx == 0.0 && ny == 0.0)
				r = 0.0;
			else
				r = atan2(nx, ny);	/* times k here is fun */
			c1 = sin(r);
			c2 = cos(r);
			t = nx;

			nx = c1 * nx - c2 * ny;
			ny = c2 * t + c1 * ny;
			break;
		}
	case 6:
		{
			/* drape */
			double      t;

			/* Avoid atan2: DOMAIN error message */
			if (nx == 0.0 && ny == 0.0)
				t = 0.0;
			else
				t = atan2(nx, ny) / M_PI;

			if (nx > 1e4 || nx < -1e4 || ny > 1e4 || ny < -1e4)
				ny = 1e4;
			else
				ny = sqrt(nx * nx + ny * ny) - 1.0;
			nx = t;
			break;
		}
	}


	/* how to check nan too?  some machines don't have finite().
	   don't need to check ny, it'll propogate */
	if (nx > 1e4 || nx < -1e4) {
		nx = halfrandom(dp, 1000) / 500.0 - 1.0;
		ny = halfrandom(dp, 1000) / 500.0 - 1.0;
		dp->fuse = FUSE;
	}
	dp->x = nx;
	dp->y = ny;
	dp->c = nc;

}

static void
draw1(driftstruct * dp)
{
	double      x = dp->x;
	double      y = dp->y;
	int         fixed_x, fixed_y, i2;

	if (dp->fuse) {
		dp->fuse--;
		return;
	}
	if (!(x > -1.0 && x < 1.0 && y > -1.0 && y < 1.0))
		return;

	fixed_x = (int) ((dp->width / 2) * (x + 1.0));
	fixed_y = (int) ((dp->height / 2) * (y + 1.0));


	dp->pts[dp->npoints].x = fixed_x;
	dp->pts[dp->npoints].y = fixed_y;
	dp->npoints++;
	if (dp->npoints == MAXBATCH1) {
		//			XSetForeground(display, gc, dp->pixcol);
		//			XDrawPoints(display, d, gc, dp->pts, dp->npoints, CoordModeOrigin);

		for(i2 = 0; i2 < dp->npoints; i2++) {
			Rectangle r;
			Point p = Pt(screen->r.min.x, screen->r.min.y);
			p = addpt(p, dp->pts[i2]);
			r = Rpt(p, addpt(p, Pt(1,1)));
			draw(screen, r, colors[dp->pixcol], nil, ZP);
		}

		dp->npoints = 0;
	}
}

static void
draw_flush(driftstruct * dp)
{
	int i2;


	if (dp->npoints)
		//			XSetForeground(display, gc, dp->pixcol);
		//		XDrawPoints(display, d, gc, dp->pts,
		//			    dp->npoints, CoordModeOrigin);

		for(i2 = 0; i2 < dp->npoints; i2++) {
			Rectangle r;
			Point p = Pt(screen->r.min.x, screen->r.min.y);

			p = addpt(p, dp->pts[i2]);
			r = Rpt(p, addpt(p, Pt(1,1)));
			draw(screen, r, colors[dp->pixcol], nil, ZP);

		}


	dp->npoints = 0;

	flushimage(display, 1);
}


void
draw_drift(void)
{
	driftstruct *dp;

	if (drifts == NULL)
		return;
	dp = &drifts[0];
	if (dp->ncpoints == NULL)
		return;

	dp->timer = 3000;
	while (dp->timer) {
		iter(dp);
		draw1(dp);
		if (dp->total_points++ > dp->fractal_len) {
			draw_flush(dp);
			if (0 == --dp->nfractals) {
				sleep(4000); /* #### make settable */
				draw(screen, screen->r, display->black, nil, ZP);
			}
			initfractal();
		}
		dp->timer--;
	}
	if (!dp->grow) {
		int         i, j, k;

		draw_flush(dp);
		if (dp->liss)
			dp->liss_time++;
		for (i = 0; i < dp->nxforms; i++)
			for (j = 0; j < 2; j++)
				for (k = 0; k < 3; k++) {
					if (dp->liss)
						dp->f[j][k][i] = sin(dp->liss_time * dp->df[j][k][i]);
					else {
						double      t = dp->f[j][k][i] += dp->df[j][k][i];

						if (t < -1.0 || 1.0 < t)
							dp->df[j][k][i] *= -1.0;
					}
				}
	}
}

void
release_drift(void)
{
	if (drifts != NULL) {

		free_drift(&drifts[0]);
		free((void *) drifts);
		drifts = (driftstruct *) NULL;
	}
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(drifts != nil)
		release_drift();
	init_drift();
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "drift") < 0)
		sysfatal("initdraw failed: %r");


	smooth(colors);
	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_drift();
		frames++;
	}
}
