/* Eruption, Copyright (c) 2002-2003 W.P. van Paassen <peter@paassen.tmfweb.nl>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.	No representations are made about the suitability of this
 * software for any purpose.	It is provided "as is" without express or
 * implied warranty.
 *
 * Module - "eruption.c"
 *
 * [02-2003] - W.P. van Paassen: Improvements, added some code of jwz from the pyro hack for a spherical distribution of the particles
 * [01-2003] - W.P. van Paassen: Port to X for use with XScreenSaver, the shadebob hack by Shane Smit was used as a template
 * [04-2002] - W.P. van Paassen: Creation for the Demo Effects Collection (http://demo-effects.sourceforge.net)
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 08/03
 */

/* plan9-related stuff */
#include "plan9.h"


/* Slightly whacked, for better explosions
 */
#define PI_2000 6284
#define SPREAD 15

static int sin_cache[PI_2000];
static int cos_cache[PI_2000];

/*particle structure*/
typedef struct 
{
	short xpos, ypos, xdir, ydir;
	unsigned char colorindex;
	unsigned char dead;
}PARTICLE;

static PARTICLE *particles;
static unsigned short iWinWidth, iWinHeight;
static unsigned char **fire;
static unsigned short nParticleCount;
static unsigned char xdelta, ydelta, decay;
static signed char gravity;
static signed short heat;

static void
cache(void) /* jwz */
{							 /*needs to be run once. Could easily be */
	int i;				/*reimplemented to run and cache at compile-time,*/
	double dA;
	for (i=0; i<PI_2000; i++)
	{
		dA=sin(((double) (random() % (PI_2000/2)))/1000.0);
		/*Emulation of spherical distribution*/
		dA+=asin(frand())/M_PI_2*0.1;
		/*Approximating the integration of the binominal, for
				well-distributed randomness*/
		cos_cache[i]=-abs((int) (cos(((double)i)/1000.0)*dA*ydelta));
		sin_cache[i]=(int) (sin(((double)i)/1000.0)*dA*xdelta);
	}
}

void init_particle(PARTICLE* particle, signed short iColorCount, unsigned short xcenter, unsigned short ycenter)
{
	int v = random() % PI_2000;
	particle->xpos = xcenter - SPREAD + (random() % (SPREAD * 2));
	particle->ypos = ycenter - SPREAD + (random() % (SPREAD * 2));
	;
	particle->xdir = sin_cache[v];
	particle->ydir = cos_cache[v];
	particle->colorindex = iColorCount;
	particle->dead = 0;
}

static void Execute( 
signed short iColorCount, uchar*aiColorVals )
{
	int i, j;
	unsigned int temp;

	/* move and draw particles into fire array */

	for (i = 0; i < nParticleCount; i++)
	{
		if (!particles[i].dead)
		{
			particles[i].xpos += particles[i].xdir;
			particles[i].ypos += particles[i].ydir;

			/* is particle dead? */

			if (particles[i].colorindex == 0)
			{
				particles[i].dead = 1;
				continue;
			}

			if (particles[i].xpos < 1)
			{
				particles[i].xpos = 1;
				particles[i].xdir = -particles[i].xdir - 4;
				particles[i].colorindex = iColorCount;
			}
			else if (particles[i].xpos >= iWinWidth - 2)
			{
				particles[i].xpos = iWinWidth - 2;
				particles[i].xdir = -particles[i].xdir + 4;
				particles[i].colorindex = iColorCount;
			}

			if (particles[i].ypos < 1)
			{
				particles[i].ypos = 1;
				particles[i].ydir = -particles[i].ydir;
				particles[i].colorindex = iColorCount;
			}
			else if (particles[i].ypos >= iWinHeight - 3)
			{
				particles[i].ypos = iWinHeight- 3;
				particles[i].ydir = (-particles[i].ydir >> 2) - (random() % 2);
				particles[i].colorindex = iColorCount;
			}

			/* gravity kicks in */
			particles[i].ydir += gravity;

			/* particle cools off */
			particles[i].colorindex--;

			/* draw particle */
			fire[particles[i].ypos][particles[i].xpos] = particles[i].colorindex;
			fire[particles[i].ypos][particles[i].xpos - 1] = particles[i].colorindex;
			fire[particles[i].ypos + 1][particles[i].xpos] = particles[i].colorindex;
			fire[particles[i].ypos - 1][particles[i].xpos] = particles[i].colorindex;
			fire[particles[i].ypos][particles[i].xpos + 1] = particles[i].colorindex;
		}
	}

	/* create fire effect */
	for (i = 0; i < iWinHeight; i++)
	{
		for (j = 0; j < iWinWidth; j++)
		{
			if (j + 1 >= iWinWidth)
				temp = 0;
			else
				temp = fire[i][j + 1];

			if (j - 1 >= 0)
				temp += fire[i][j - 1];

			if (i - 1 >= 0)
			{
				temp += fire[i - 1][j];
				if (j - 1 >= 0)
					temp += fire[i - 1][j - 1];
				if (j + 1 < iWinWidth)
					temp += fire[i - 1][j + 1];
			}

			if (i + 1 < iWinHeight)
			{
				temp += fire[i + 1][j];
				if (j + 1 < iWinWidth)
					temp += fire[i + 1][j + 1];
				if (j - 1 >= 0)
					temp += fire[i + 1][j - 1];
			}

			temp >>= 3;

			if (temp > decay)
			{
				temp -= decay;
			}
			else {
				temp = 0;
				if(fire[i][j] != 0) {
					buf[i*iWinWidth + j] = aiColorVals[ 0 ];
				}
			}
			fire[i][j] = temp;
		}
	}

	//memset( pImage->data, 0, pImage->bytes_per_line * pImage->height );

	/* draw fire array to screen */
	for (i = 0; i < iWinHeight; ++i)
	{
		for (j = 0; j < iWinWidth; ++j)
		{
			if (fire[i][j] > 0) {
				//XPutPixel( pImage, j, i, aiColorVals[ fire[i][j] ] );

				/* plan 9 baby! */
				//r.max.x = (r.min.x = screen->r.min.x + j) + 1;
				//r.max.y = (r.min.y = screen->r.min.y + i) + 1;
				//draw(screen, r, aiColorVals[ fire[i][j] ], nil, ZP);
				buf[i*iWinWidth + j] = aiColorVals[ fire[i][j] ];

			}

		}
	}
	/*	XPutImage( pDisplay, MainWindow, *pGC, pImage,
 *			 0,0,0,0, iWinWidth, iWinHeight );
 *	XSync( pDisplay, False );
 */
	loadimage(img, Rpt(Pt(0, 0), Pt(Dx(screen->r), Dy(screen->r))), buf, iWinWidth*(iWinHeight));
	draw(screen, screen->r, img, nil, ZP);
	flushimage(display, 1);
}

static uchar* SetPalette(signed short *piColorCount )
{
	XColor Color, *aColors;
	uchar *aiColorVals;
	signed short iColor;

	*piColorCount = 256;

	aColors		= calloc( *piColorCount, sizeof(XColor) );
	aiColorVals = calloc( *piColorCount, sizeof(uchar) );

	Color.red = Color.green = Color.blue = 65535 / *piColorCount;

	/* create fire palette */
	for( iColor=0; iColor < *piColorCount; iColor++ )
	{
		if (iColor < *piColorCount >> 3)
		{
			/* black to blue */
			aColors[iColor].red = 0;
			aColors[iColor].green = 0;
			aColors[iColor].blue = Color.blue * (iColor << 1);
		}
		else if (iColor < *piColorCount >> 2)
		{
			/* blue to red */
			signed short temp = (iColor - (*piColorCount >> 3));
			aColors[iColor].red = Color.red * (temp << 3);
			aColors[iColor].green = 0;
			aColors[iColor].blue = 16383 - Color.blue * (temp << 1);
		}
		else if (iColor < (*piColorCount >> 2) + (*piColorCount >> 3))
		{
			/* red to yellow */
			signed short temp = (iColor - (*piColorCount >> 2)) << 3;
			aColors[iColor].red = 65535;
			aColors[iColor].green = Color.green * temp;
			aColors[iColor].blue = 0;
		}
		else if (iColor < *piColorCount >> 1)
		{
			/* yellow to white */
			signed int temp = (iColor - ((*piColorCount >> 2) + (*piColorCount >> 3))) << 3;
			aColors[iColor].red = 65535;
			aColors[iColor].green = 65535;
			aColors[iColor].blue = Color.blue * temp;
		}
		else
		{
			/* white */
			aColors[iColor].red = aColors[iColor].green = aColors[iColor].blue = 65535;
		}

		aiColorVals[ iColor ] = rgb2cmap(aColors[ iColor ].red, 
			aColors[ iColor ].green, 
			aColors[ iColor ].blue);
	}

	if (heat < *piColorCount)
		*piColorCount = heat;

	free( aColors );

	//XSetWindowBackground( pDisplay, Win, aiColorVals[ 0 ] );
	//draw(screen, screen->r, aiColorVals[ 0 ], nil, ZP);

	return aiColorVals;
}


static void Initialize( void )
{

	/* Create the Image for drawing */
	int i;

	iWinWidth = Dx(screen->r);
	iWinHeight = Dy(screen->r);

	/* create fire array */
	fire = calloc( iWinHeight, sizeof(unsigned char*));
	for (i = 0; i < iWinHeight; ++i) {
		fire[i] = calloc( iWinWidth, sizeof(unsigned char));
	}

	/*create particles */
	particles = malloc (nParticleCount * sizeof(PARTICLE));
}

void screenhack( void )
{
	signed short iColorCount = 0;
	uchar*aiColorVals = NULL;
	unsigned short sum = 0;
	int delay, cycles, i;

	nParticleCount = 100;	/* 100 <-> 2000 */
	decay = 2;				/* 0<->10 */
	gravity = 1;			/* -5 <-> 5 */
	heat = 256;			/* 64 <-> 256 */


	Initialize(	);

	ydelta = 0;
	while (sum < (iWinHeight >> 1) - SPREAD)
	{
		ydelta++;
		sum += ydelta;
	}

	sum = 0;
	while (sum < (iWinWidth >> 3))
	{
		xdelta++;
		sum += xdelta;
	}

	delay = 0;
	cycles = 80;
	i = cycles;

	cache();

	free( aiColorVals );
	aiColorVals = SetPalette( &iColorCount );
	//draw(screen, screen->r, aiColorVals[0], nil, ZP);

	while( 1 )
	{

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}

		if( i++ >= cycles )
		{
			/* compute random center */
			unsigned short xcenter, ycenter;
			xcenter = random() % iWinWidth;
			ycenter = random() % iWinHeight;

			for (i = 0; i < nParticleCount; i++)
				init_particle(particles + i, iColorCount - 1, xcenter, ycenter);
			i = 0;
		}
		Execute( iColorCount - 1, aiColorVals );
		frames++;
		if( delay && !(i % 4) )
			sleep(delay);

	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	fprint(2, "resize not supported...\n");
	threadexitsall("resize not supported");

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "eruption") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	buf = malloc(Dx(screen->r)*Dy(screen->r));
	if(buf == nil)
		sysfatal("cannot malloc buffer: %r\n");

	img = allocimage(display, Rpt(Pt(0, 0), Pt(Dx(screen->r), Dy(screen->r))), CMAP8, 1, DCyan);
	if(img == nil)
		sysfatal("cannot allocate image: %r\n");

	screenhack();
}
