/* fadeplot --- a fading plot of sine squared */


/*-
 * Some easy plotting stuff, by Bas van Gaalen, Holland, PD
 *
 * Copyright (c) 1996 by Charles Vidal
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 10-May-1997: Compatible with screensaver
 * 1996: Written by Charles Vidal based on work by Bas van Gaalen
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 07/02
  * todo: it just doesn't look good...
  */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"
#define CYCLES 1500
#define COUNT 10
/* end of plan9-related defines */


#define MINSTEPS 1

typedef struct {
	XPoint      speed, step, factor, st;
	int         temps, maxpts, nbstep;
	int         min;
	int         width, height;
	int         pix;
	int         angles;
	int        *stab;
	XPoint     *pts;
} fadeplotstruct;

static fadeplotstruct *fadeplots = (fadeplotstruct *) NULL;

static void
free_fadeplot(fadeplotstruct *fp)
{
	if (fp->pts != NULL) {
		free((void *) fp->pts);
		fp->pts = (XPoint *) NULL;
	}
	if (fp->stab != NULL) {
		free((void *) fp->stab);
		fp->stab = (int *) NULL;
	}
}

static Bool
initSintab(void)
{
	fadeplotstruct *fp = &fadeplots[0];
	int         i;
	float       x;

	fp->angles = NRAND(950) + 250;
	if ((fp->stab = (int *) malloc(fp->angles * sizeof (int))) == NULL) {
		free_fadeplot(fp);
		return False;
	}
	for (i = 0; i < fp->angles; i++) {
		x = sin(2.0 * M_PI * i / fp->angles);
		fp->stab[i] = (int) (x * (x > 0 ? x : -x) * fp->min) + fp->min;
	}
	return True;
}

void
init_fadeplot(void)
{
	fadeplotstruct *fp;

	if (fadeplots == NULL) {
		if ((fadeplots = (fadeplotstruct *) calloc(1,
			sizeof (fadeplotstruct))) == NULL)
			return;
	}
	fp = &fadeplots[0];

	fp->width = Dx(screen->r);
	fp->height = Dy(screen->r);
	fp->min = MAX(MIN(fp->width, fp->height) / 2, 1);

	fp->speed.x = 8;
	fp->speed.y = 10;
	fp->step.x = 1;
	fp->step.y = 1;
	fp->temps = 0;
	fp->factor.x = MAX(fp->width / (2 * fp->min), 1);
	fp->factor.y = MAX(fp->height / (2 * fp->min), 1);

	fp->nbstep = COUNT;

	fp->maxpts = CYCLES;

	if (fp->pts == NULL) {
		if ((fp->pts = (XPoint *) calloc(fp->maxpts, sizeof (XPoint))) ==
			NULL) {
			free_fadeplot(fp);
			return;
		}
	}
	fp->pix = NRAND(255);

	if (fp->stab != NULL)
		free((void *) fp->stab);
	if (!initSintab())
		return;
	draw(screen, screen->r, display->black, nil, ZP);
}

void
draw_fadeplot(void)
{
	int         i, i2, j, temp;
	fadeplotstruct *fp;

	if (fadeplots == NULL)
		return;
	fp = &fadeplots[0];
	if (fp->stab == NULL)
		return;

	//	XSetForeground(display, gc, MI_BLACK_PIXEL(mi));
	//	XDrawPoints(display, window, gc, fp->pts, fp->maxpts, CoordModeOrigin);
	/*		for(i2 = 0; i2 < fp->maxpts; i2++) {
			Rectangle r;
			Point p = addpt(screen->r.min, fp->pts[i2]);

			r = Rpt(p, addpt(p, Pt(1, 1)));
			draw(screen, r, display->black, nil, ZP);
		}
*/
	draw(screen, screen->r, display->black, nil, ZP);
	flushimage(display, 1);

	if (++fp->pix >= 255)
		fp->pix = 1;

	temp = 0;
	for (j = 0; j < fp->nbstep; j++) {
		for (i = 0; i < fp->maxpts / fp->nbstep; i++) {
			fp->pts[temp].x =
				fp->stab[(fp->st.x + fp->speed.x * j + i * fp->step.x) % fp->angles] *
				fp->factor.x + fp->width / 2 - fp->min;
			fp->pts[temp].y =
				fp->stab[(fp->st.y + fp->speed.y * j + i * fp->step.y) % fp->angles] *
				fp->factor.y + fp->height / 2 - fp->min;
			temp++;
		}
	}
	//	XDrawPoints(display, window, gc, fp->pts, temp, CoordModeOrigin);
	for(i2 = 0; i2 < temp; i2++) {
		Rectangle r;
		Point p = addpt(screen->r.min, fp->pts[i2]);

		r = Rpt(p, addpt(p, Pt(1, 1)));
		draw(screen, r, colors[fp->pix], nil, ZP);
	}

	fp->st.x = (fp->st.x + fp->speed.x) % fp->angles;
	fp->st.y = (fp->st.y + fp->speed.y) % fp->angles;
	fp->temps++;
	if ((fp->temps % (fp->angles / 2)) == 0) {
		fp->temps = fp->temps % fp->angles * 5;
		if ((fp->temps % (fp->angles)) == 0)
			fp->speed.y = (fp->speed.y++) % 30 + 1;
		if ((fp->temps % (fp->angles * 2)) == 0)
			fp->speed.x = (fp->speed.x) % 20;
		if ((fp->temps % (fp->angles * 3)) == 0)
			fp->step.y = (fp->step.y++) % 2 + 1;

		draw(screen, screen->r, display->black, nil, ZP);
	}
	flushimage(display, 1);
}

void
release_fadeplot(void)
{
	if (fadeplots != NULL) {

		free_fadeplot(&fadeplots[0]);
		free((void *) fadeplots);
		fadeplots = (fadeplotstruct *) NULL;
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(fadeplots != nil)
		release_fadeplot();
	init_fadeplot();
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "fadeplot") < 0)
		sysfatal("initdraw failed: %r");

	smooth(colors);
	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_fadeplot();
		frames++;
	}
}
