/* forest.c (aka xtree.c), Copyright (c) 1999
 *  Peter Baumung <unn6@rz.uni-karlsruhe.de>
 *
 * Most code taken from
 *  xscreensaver, Copyright (c) 1992, 1995, 1997 
 *  Jamie Zawinski <jwz@netscape.com>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
  * ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 09/03
  */

/* plan9-related stuff */
#include "plan9.h"

static int col;


typedef struct {
	int     x;
	int     y;
	int     thick;
	double  size;
	long    color;
	int     toDo;
	int     season;
} treestruct;


static treestruct *trees = NULL;


static long colorM[12] = {
	0xff0000, 0xff8000, 0xffff00, 0x80ff00,
	0x00ff00, 0x00ff80, 0x00ffff, 0x0080ff,
	0x0000ff, 0x8000ff, 0xff00ff, 0xff0080};

static long colorV[12] = {
	0x0a0000, 0x0a0500, 0x0a0a00, 0x050a00,
	0x000a00, 0x000a05, 0x000a0a, 0x00050a,
	0x00000a, 0x05000a, 0x0a000a, 0x0a0005};

void init_trees(void) {
	treestruct       *tree;
	int 			i;

	if (trees == NULL) {
		trees = (treestruct *) calloc(1, sizeof (treestruct));
		if (trees == NULL) {
			return;
		}
	}


	draw(screen, screen->r, display->black, nil, ZP);

	tree = trees;
	tree->toDo   = 25;
	tree->season = NRAND(12);
	for (i = 4; i < 20; i++) {
		ulong color, tmpcol;
		int sIndex = (tree->season + (i-4) / 4) % 11 + 1;

		color = colorM[sIndex] - 2 * colorV[sIndex] * (i % 4);
		tmpcol = ((color & 0xff0000) / 256) << 24 |
			(color & 0x00ff00) << 16 | 
			((color & 0x0000ff) * 256) << 8  | 0xFF;


		freeimage(colors[i]);
		//print("%.2d: sindex: %d; color: %8.8ulx\n", i, sIndex, tmpcol);

		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, tmpcol);
	}

	col = 1;
}

double rRand(double a, double b) {
	return (a+(b-a)*NRAND(10001)/10000.0);
}

void draw_line(int x1, int y1, int x2, int y2,
double angle, int widths, int widthe) {

	double      sns = 0.5*widths*sin(angle + M_PI_2);
	double      css = 0.5*widths*cos(angle + M_PI_2);
	double      sne = 0.5*widthe*sin(angle + M_PI_2);
	double      cse = 0.5*widthe*cos(angle + M_PI_2);

	int         xs1 = (int) (x1-sns);
	int         xs2 = (int) (x1+sns);
	int         ys1 = (int) (y1-css);
	int         ys2 = (int) (y1+css);
	int         xe1 = (int) (x2-sne);
	int         xe2 = (int) (x2+sne);
	int         ye1 = (int) (y2-cse);
	int         ye2 = (int) (y2+cse);
	int         i;

	for (i = 0; i < widths; i++) {
		if (col >= 4) {
			//XSetForeground(display, gc, colors[i*4/widths].pixel);
			col = i*4/widths;
		}

		//		XDrawLine(display, MI_WINDOW(mi), gc,
		//		    xs1+(xs2-xs1)*i/widths, ys1+(ys2-ys1)*i/widths,
		//		    xe1+(xe2-xe1)*i/widths, ye1+(ye2-ye1)*i/widths);

		line(screen, addpt(screen->r.min, Pt(xs1+(xs2-xs1)*i/widths, ys1+(ys2-ys1)*i/widths)), 
			addpt(screen->r.min, Pt(xe1+(xe2-xe1)*i/widths, ye1+(ye2-ye1)*i/widths)), 
			Endsquare, Endsquare, 0, colors[col], ZP);

	}
}

void draw_tree_rec(double thick, int x, int y, double angle) {
	treestruct *tree = trees;
	int         length = (24+NRAND(12))*tree->size;
	int         a = (int) (x - length*sin(angle));
	int         b = (int) (y - length*cos(angle));
	int         i;

	draw_line(x, y, a, b, angle, thick*tree->size, 0.68*thick*tree->size);

	if (thick > 2) {
		draw_tree_rec(0.68*thick, a, b, 0.8*angle+rRand(-0.2, 0.2));
		if (thick < tree->thick-1) {
			draw_tree_rec(0.68*thick, a, b, angle+rRand(0.2, 0.9));
			draw_tree_rec(0.68*thick, (a+x)/2, (b+y)/2, angle-rRand(0.2, 0.9));
		}
	}

	if (thick < 0.5*tree->thick) {
		int     nleaf = 12 + NRAND(4);
		XArc    leaf[16];
		for (i = 0; i < nleaf; i++) {
			leaf[i].x = a + (int) (tree->size * rRand(-12, 12));
			leaf[i].y = b + (int) (tree->size * rRand(-12, 12));
			leaf[i].width = (int) (tree->size * rRand(2, 6));
			leaf[i].height = leaf[i].width;
			leaf[i].angle1 = 0;
			leaf[i].angle2 = 360 * 64;
		}

		col= tree->color +NRAND(4);

		//XFillArcs(display, MI_WINDOW(mi), gc, leaf, nleaf);
		for(i = 0; i < nleaf; i++) {
			fillellipse(screen, addpt(screen->r.min, 
				Pt(leaf[i].x,leaf[i].y)), 
				leaf[i].width/2, leaf[i].height/2, colors[col], ZP);
		}
	}
	flushimage(display, 1);
}

void draw_trees(void) {
	treestruct *tree = trees;
	int		width = Dx(screen->r);
	int		height = Dy(screen->r);

	if (--(tree->toDo) == 0) {
		sleep(3000);
		init_trees();
	}

	tree->x        = NRAND(width);
	tree->y        = (int) (1.25 * height * (1 - tree->toDo / 23.0));
	tree->thick    = rRand(7, 12);
	tree->size     = height / 480.0;
	tree->color    = 4 * (NRAND(4) + 1);
	draw_tree_rec(tree->thick, tree->x, tree->y, rRand(-0.1, 0.1));
}

void release_trees(void) {
	if (trees != NULL) {
		free((void *) trees);
		trees = NULL;
	}
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(trees != nil)
		release_trees();
	init_trees();
}


void 
threadmain(int argc, char **argv)
{
	Mouse m;
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "forest") < 0)
		sysfatal("initdraw failed: %r");


	for(i = 0; i < 4; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			(24576 + 4096 * (i % 4)) << 24 |
			(24576 + 2048 * (i % 4)) << 16 | 0x00 << 8  | 0xFF);
	}
	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		frames++;
		draw_trees();
	}
}
