/* xscreensaver, Copyright (c) 1992, 1995, 1996, 1997
 *  Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* Algorithm from a Mac program by Chris Tate, written in 1988 or so. */

/* 18-Sep-97: Johannes Keukelaar (johannes@nada.kth.se): Improved screen
 *            eraser.
 * 10-May-97: merged ellipse code by Dan Stromberg <strombrg@nis.acs.uci.edu>
 *            as found in xlockmore 4.03a10.
 * 1992:      jwz created.
 */

/* 25 April 2002: Matthew Strait <straitm@mathcs.carleton.edu> added
-subdelay option so the drawing process can be watched */

/*
  * Ported to Plan 9 by andrey@lanl.gov, 07/02
  */


/* plan9-related stuff */
#include "plan9.h"

static int color_index;

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

}


static double sins [360];
static double coss [360];

static unsigned int default_fg_pixel;
static int sleep_time;
static int subdelay;

static void
init_helix (void)
{
	int i;

	for (i = 0; i < 360; i++)
	{
		sins [i] = sin ((((double) i) / 180.0) * M_PI);
		coss [i] = cos ((((double) i) / 180.0) * M_PI);
	}
}

static int
gcd (int a, int b)
{
	while (b > 0)
	{
		int tmp;
		tmp = a % b;
		a = b;
		b = tmp;
	}
	return (a < 0 ? -a : a);
}

static void
helix (
int radius1, int radius2, int d_angle,
int factor1, int factor2, int factor3, int factor4)
{
	int width, height;
	int xmid, ymid;
	int x1, y1, x2, y2, angle, limit;
	int i;

	draw(screen, screen->r, display->black, nil, ZP);


	width = Dx(screen->r);
	height = Dy(screen->r);

	xmid = width / 2;
	ymid = height / 2;
	x2 = xmid;
	y2 = ymid + radius1;
	angle = 0;
	limit = 1 + (360 / gcd (360, d_angle));

	for (i = 0; i < limit; i++)
	{
		int tmp;
#define pmod(x,y) (tmp=((x) % (y)), (tmp >= 0 ? tmp : (tmp + (y))))

		x1 = xmid + (((double) radius1) * sins [pmod ((angle * factor1), 360)]);
		y1 = ymid + (((double) radius2) * coss [pmod ((angle * factor2), 360)]);
		//      XDrawLine (dpy, window, draw_gc, x1, y1, x2, y2);

		line(screen, Pt(screen->r.min.x +x1,	screen->r.min.y +y1), 
			Pt(screen->r.min.x +x2, screen->r.min.y +y2),
			Endsquare, Endsquare, 0, colors [color_index], ZP);


		x2 = xmid + (((double) radius2) * sins [pmod ((angle * factor3), 360)]);
		y2 = ymid + (((double) radius1) * coss [pmod ((angle * factor4), 360)]);
		//      XDrawLine (dpy, window, draw_gc, x1, y1, x2, y2);

		line(screen, Pt(screen->r.min.x +x1, screen->r.min.y +y1), 
			Pt(screen->r.min.x +x2, screen->r.min.y +y2),
			Endsquare, Endsquare, 0, colors [color_index], ZP);

		angle += d_angle;

		/* if we sleep every time, it's too slow */
		if(subdelay && i%16 == 0) sleep(subdelay);

		flushimage(display, 1);
	}
}

static void
trig (
int d_angle, int factor1, int factor2,
int offset, int d_angle_offset, int dir, int density)
{
	int width, height;
	int xmid, ymid;
	int x1, y1, x2, y2;
	int tmp, angle;

	draw(screen, screen->r, display->black, nil, ZP);
	width = Dx(screen->r);
	height = Dy(screen->r);

	xmid = width / 2;
	ymid = height / 2;

	while (d_angle >= -360 && d_angle <= 360)
	{
		angle = d_angle + d_angle_offset;
		x1 = (sins [pmod(angle * factor1, 360)] * xmid) + xmid;
		y1 = (coss [pmod(angle * factor1, 360)] * ymid) + ymid;
		x2 = (sins [pmod(angle * factor2 + offset, 360)] * xmid) + xmid;
		y2 = (coss [pmod(angle * factor2 + offset, 360)] * ymid) + ymid;
		//      XDrawLine(dpy, window, draw_gc, x1, y1, x2, y2);

		line(screen, Pt(screen->r.min.x +x1, screen->r.min.y +y1), 
			Pt(screen->r.min.x +x2, screen->r.min.y +y2),
			Endsquare, Endsquare, 0, colors [color_index], ZP);


		tmp = (int) 360 / (2 * density * factor1 * factor2);
		if (tmp == 0)	/* Do not want it getting stuck... */
			tmp = 1;	/* Would not need if floating point */
		d_angle += dir * tmp;

		/* this draws faster, so we sleep somewhat more often */
		if(subdelay && d_angle%4 == 0) sleep(subdelay);

		/* without this, the subdelay effect is lost */
		flushimage(display, 1);
	}
}


static void
random_helix (void)
{
	int width, height;
	int radius, radius1, radius2, d_angle, factor1, factor2, factor3, factor4;
	double divisor;

	width = Dx(screen->r);
	height = Dy(screen->r);

	radius = MIN(width, height) / 2;

	d_angle = 0;
	factor1 = 2;
	factor2 = 2;
	factor3 = 2;
	factor4 = 2;

	divisor = ((frand () * 3.0 + 1) * (((random() & 1) * 2) - 1));

	if ((random () & 1) == 0)
	{
		radius1 = radius;
		radius2 = radius / divisor;
	}
	else
	{
		radius2 = radius;
		radius1 = radius / divisor;
	}

	while (gcd (360, d_angle) >= 2)
		d_angle = random () % 360;

#define random_factor()				\
  (((random() % 7) ? ((random() & 1) + 1) : 3)	\
   * (((random() & 1) * 2) - 1))

	while (gcd (gcd (gcd (factor1, factor2), factor3), factor4) != 1)
	{
		factor1 = random_factor ();
		factor2 = random_factor ();
		factor3 = random_factor ();
		factor4 = random_factor ();
	}


	helix (radius1, radius2, d_angle,
		factor1, factor2, factor3, factor4);
}

static void
random_trig (void)
{
	int d_angle, factor1, factor2;
	int offset, d_angle_offset, dir, density;



	d_angle = 0;
	factor1 = (random() % 8) + 1;
	do 
	{
		factor2 = (random() % 8) + 1;
	}	while (factor1 == factor2);

	dir = (random() & 1) ? 1 : -1;
	d_angle_offset = random() % 360;
	offset = ((random() % ((360 / 4) - 1)) + 1) / 4;
	density = 1 << ((random() % 4) + 4);	/* Higher density, higher angles */


	trig (d_angle, factor1, factor2,
		offset, d_angle_offset, dir, density);
}

static void
random_helix_or_trig (void)
{

	if (random() & 1)
		random_helix();
	else
		random_trig();

	if(ecanmouse()) {
		m = emouse();
		if(m.buttons&4)
			if(emenuhit(3, &m, &menu) == 0) {
				threadexitsall(0);
			}
	}
	sleep ( sleep_time );

	if(ecanmouse()) {
		m = emouse();
		if(m.buttons&4)
			if(emenuhit(3, &m, &menu) == 0) {
				threadexitsall(0);
			}
	}
	draw(screen, screen->r, display->black, nil, ZP);
	color_index = nrand(255) + 1;
	frames++;
}


void
screenhack (void)
{
	sleep_time = 500;
	subdelay = 10;
	init_helix ();
	while (1)
		random_helix_or_trig ();
}




void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "helix") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(i<<2));
	}

	einit(Emouse);

	color_index = nrand(255) + 1;

	screenhack();
}
