/* imsmap, Copyright (c) 1992 Juergen Nickelsen <nickel@cs.tu-berlin.de>
 * Derived from code by Markus Schirmer, TU Berlin.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * Revision History:
 * 24-aug-92: jwz: hacked.
 * 17-May-97: jwz: hacked more.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"
#include "smooth.h"

#define NSTEPS 7
#define COUNT (1 << NSTEPS)
#define CELL(c, r) cell[((unsigned int)(c)) + ((unsigned int) (r)) * xmax]

static enum imsmap_mode {
	MODE_H, MODE_S, MODE_V, MODE_RANDOM } mode;


static int ncolors = 0;
static int cycle_direction;
static Bool extra_krinkly_p;

static int delay, cycle_delay;
static signed char *cell = NULL;
static int xmax, ymax;
static int iterations;


static void
init_map (void)
{
	static Bool rv_p;


	if (!ncolors)
	{
		rv_p = nrand(2);
		ncolors = 256;
		delay = 1;
		cycle_delay = 100;
		iterations = 7;


		mode = MODE_RANDOM;
	}


	extra_krinkly_p = !(random() % 5);

	smooth(colors);

	//XSetForeground (dpy, gc, colors[1].pixel);
	//XFillRectangle (dpy, window, gc, 0, 0, xgwa.width, xgwa.height);
	draw(screen, screen->r, colors[0], nil, ZP);
}


#define HEIGHT_TO_PIXEL(height)				\
	((height) < 0					\
	 ? (extra_krinkly_p				\
	    ? ncolors - ((-(height)) % ncolors)		\
	    : 0)					\
	 : ((height) >= ncolors				\
	    ? (extra_krinkly_p				\
	       ? (height) % ncolors			\
	       : ncolors-1)				\
	    : (height)))


static unsigned int
set (unsigned int l,
unsigned int c,
unsigned int size,
int height)
{
	int rang = 1 << (NSTEPS - size);
	height = height + (random () % rang) - rang / 2;
	height = HEIGHT_TO_PIXEL(height);
	CELL (l, c) = height;
	return height;
}

static void
floyd_steinberg (void)
{
	int x, y, err;

	/* Instead of repeatedly calling XPutPixel(), we make an Image and then
     send its bits over all at once.  This consumes much less network
     bandwidth.  The image we create is Wx1 intead of WxH, so that we
     don't use enormous amounts of memory.
   */
	//	XImage *image =
	//	    XCreateImage (dpy, xgwa.visual,
	//	    1, XYBitmap, 0,		/* depth, format, offset */
	//	(char *) calloc ((xmax + 8) / 8, 1),	/* data */
	//	xmax, 1, 8, 0);		/* w, h, pad, bpl */

	//	XSetForeground (dpy, gc, colors[0].pixel);
	//	XSetBackground (dpy, gc, colors[1].pixel);

	for (y = 0; y < ymax - 1; y++)
	{
		for (x = 0; x < xmax - 1; x++)
		{
			if (CELL(x, y) < 0)
			{
				err = CELL (x, y);
				//XPutPixel (image, x, 0, 1);
				draw(screen, rectaddpt(Rect(x, y, x+1, y+1), screen->r.min), colors[1], nil, ZP);
			}
			else
			{
				err = CELL (x, y) - 1;
				//XPutPixel (image, x, 0, 0);
				draw(screen, rectaddpt(Rect(x, y, x+1, y+1), screen->r.min), colors[0], nil, ZP);
			}
			/* distribute error */
			CELL (x,   y+1) += (int) (((float) err) * 3.0/8.0);
			CELL (x+1, y)   += (int) (((float) err) * 3.0/8.0);
			CELL (x+1, y+1) += (int) (((float) err) * 1.0/4.0);
		}
		//XPutImage (dpy, window, gc, image, 0, 0, 0, y, xmax, 1);
	}
	//	XDestroyImage (image);
}

static void
drawp (int x, int y, unsigned long pixel, int grid_size)
{
	static unsigned int last_pixel, last_valid = 0;
	//	if (! (last_valid && pixel == last_pixel))
	//		XSetForeground (dpy, gc, pixel);
	last_valid = 1, last_pixel = pixel;

	//XFillRectangle (dpy, window, gc, x, y, grid_size, grid_size);
	draw(screen, rectaddpt(Rect(x, y, x+grid_size, y+grid_size), screen->r.min), 
		colors[pixel], nil, ZP);
}


static void 
draw_map (void)
{
	int xstep, ystep, xnextStep, ynextStep;
	int x, y, i, x1, x2, y1, y2;
	unsigned int pixel, qpixels [4];

	int backwards = random() & 1;

	xmax = Dx(screen->r);
	ymax = Dy(screen->r);

	if (cell == NULL)
		exits ("cell is null; call eresized(0) before draw_map");

	CELL (0, 0) = 0;
	xstep = (backwards ? -COUNT : COUNT);
	ystep = COUNT;
	for (i = 0; i < iterations; i++)
	{
		xnextStep = xstep / 2;
		ynextStep = ystep / 2;
		for (x = (backwards ? xmax-1 : 0);
			(backwards ? x >= 0 : x < xmax);
			x += xstep)
		{
			x1 = x + xnextStep;
			if (x1 < 0)
				x1 = xmax-1;
			else if (x1 >= xmax)
				x1 = 0;

			x2 = x + xstep;
			if (x2 < 0)
				x2 = xmax-1;
			else if (x2 >= xmax)
				x2 = 0;

			for (y = 0; y < ymax; y += ystep)
			{
				y1 = y + ynextStep;
				if (y1 < 0)
					y1 = ymax-1;
				else if (y1 >= ymax)
					y1 = 0;

				y2 = y + ystep;
				if (y2 < 0)
					y2 = ymax-1;
				else if (y2 >= ymax)
					y2 = 0;

				qpixels [0] = HEIGHT_TO_PIXEL (CELL (x, y));
				qpixels [1] = HEIGHT_TO_PIXEL (CELL (x, y2));
				qpixels [2] = HEIGHT_TO_PIXEL (CELL (x2, y));
				qpixels [3] = HEIGHT_TO_PIXEL (CELL (x2, y2));

				pixel = set (x, y1, i,
					((int) CELL (x, y) + (int) CELL (x, y2) + 1) / 2);
				if ((pixel != qpixels[0] || pixel != qpixels[1] || 
					pixel != qpixels[2] || pixel != qpixels[3]))
					drawp (x, y1, pixel, ynextStep);

				pixel = set (x1, y, i,
					((int) CELL (x, y) + (int) CELL (x2, y) + 1) / 2);
				if ((pixel != qpixels[0] || pixel != qpixels[1] || 
					pixel != qpixels[2] || pixel != qpixels[3]))
					drawp (x1, y, pixel, ynextStep);

				pixel = set (x1, y1, i,
					((int) CELL (x, y) + (int) CELL (x, y2) +
					(int) CELL (x2, y) + (int) CELL (x2, y2) + 2)
					/ 4);
				if ((pixel != qpixels[0] || pixel != qpixels[1] || 
					pixel != qpixels[2] || pixel != qpixels[3]))
					drawp (x1, y1, pixel, ynextStep);


			}
		}
		xstep = xnextStep;
		ystep = ynextStep;
		flushimage(display, 1);
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
	}

	//XSync (dpy, False);
	flushimage(display, 1);
}


void
screenhack (void)
{
	while (1)
	{
		init_map ();
		draw_map ();
		frames++;

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if (delay)
			sleep (delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	if(cell)
		free(cell);

	cell = (signed char *) calloc (Dx(screen->r) * Dy(screen->r), 1);

	draw(screen, screen->r, display->black, nil, ZP);
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "imsmap") < 0)
		sysfatal("initdraw failed: %r");


	einit(Emouse);
	eresized(0);

	screenhack();
}
