/* mountain -- square grid mountains */


/*-
 * Copyright (c) 1995 Pascal Pensa <pensa@aurora.unice.fr>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 10-May-1997: Compatible with xscreensaver
 * 1995: Written
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 07/02
  */

/* plan9-related stuff */
#include "plan9.h"

#define CYCLES 50000
#define COUNT 50
/* end of plan9-related defines */




/* ~ 5000 Max mountain height (1000 - 10000) */
#define MAXHEIGHT  (3 * (mp->width + mp->height))

#define WORLDWIDTH 50		/* World size x * y */

#define RANGE_RAND(min,max) ((min) + NRAND((max) - (min)))

typedef struct {
	int         pixelmode;
	int         width;
	int         height;
	int         x, y;
	int         offset;
	int         stage;
	int         h[WORLDWIDTH][WORLDWIDTH];
	long        time;	/* up time */
	Bool        wireframe;
	Bool        joke;
} mountainstruct;

static mountainstruct *mountains = (mountainstruct *) NULL;

static void
spread(int  (*m)[50], int x, int y)
{
	int         x2, y2;
	int         h = m[x][y];

	for (y2 = y - 1; y2 <= y + 1; y2++)
		for (x2 = x - 1; x2 <= x + 1; x2++)
			if (x2 >= 0 && y2 >= 0 && x2 < WORLDWIDTH && y2 < WORLDWIDTH)
				m[x2][y2] = (m[x2][y2] + h) / 2;
}

static void
drawamountain(void)
{
	mountainstruct *mp = &mountains[0];
	int         x2, y2, x3, y3, y4, y5, c;
	XPoint      p[4];

	c = (mp->h[mp->x][mp->y] + mp->h[mp->x + 1][mp->y] +
		mp->h[mp->x][mp->y + 1] + mp->h[mp->x + 1][mp->y + 1]) / 4;
	c = (c / 10 + mp->offset) % 255;

	x2 = mp->x * (2 * mp->width) / (3 * WORLDWIDTH);
	y2 = mp->y * (2 * mp->height) / (3 * WORLDWIDTH);
	p[0].x = (x2 - y2 / 2) + (mp->width / 4);
	p[0].y = (y2 - mp->h[mp->x][mp->y]) + mp->height / 4;

	p[0] = addpt(screen->r.min, p[0]);

	x3 = (mp->x + 1) * (2 * mp->width) / (3 * WORLDWIDTH);
	y3 = mp->y * (2 * mp->height) / (3 * WORLDWIDTH);
	p[1].x = (x3 - y3 / 2) + (mp->width / 4);
	p[1].y = (y3 - mp->h[mp->x + 1][mp->y]) + mp->height / 4;
	p[1] = addpt(screen->r.min, p[1]);

	y4 = (mp->y + 1) * (2 * mp->height) / (3 * WORLDWIDTH);
	p[2].x = (x3 - y4 / 2) + (mp->width / 4);
	p[2].y = (y4 - mp->h[mp->x + 1][mp->y + 1]) + mp->height / 4;
	p[2] = addpt(screen->r.min, p[2]);


	y5 = (mp->y + 1) * (2 * mp->height) / (3 * WORLDWIDTH);
	p[3].x = (x2 - y5 / 2) + (mp->width / 4);
	p[3].y = (y5 - mp->h[mp->x][mp->y + 1]) + mp->height / 4;
	p[3] = addpt(screen->r.min, p[3]);


	if (mp->joke) {
		if ((Bool) (LRAND() & 1)) {
			//			XDrawLines(display, window, gc, p, 5, CoordModeOrigin);
			poly(screen, p, 4, Endsquare, Endsquare, 1, colors[c], screen->r.min);
		}
		else {
			//			XFillPolygon(display, window, gc, p, 4, Complex, CoordModeOrigin);
			fillpoly(screen, p, 4, 1, colors[c], screen->r.min);
			if (!mp->pixelmode) {
				//				XSetForeground(display, gc, MI_BLACK_PIXEL(mi));
				//				XDrawLines(display, window, gc, p, 5, CoordModeOrigin);
				poly(screen, p, 4, Endsquare, Endsquare, 1, display->black, screen->r.min);
			}
		}
	}
	else {
		if (mp->wireframe) {
			//			XDrawLines(display, window, gc, p, 5, CoordModeOrigin);
			poly(screen, p, 4, Endsquare, Endsquare, 1, colors[c], screen->r.min);
		}
		else {
			//			XFillPolygon(display, window, gc, p, 4, Complex, CoordModeOrigin);

			fillpoly(screen, p, 4, 1, colors[c], screen->r.min);
			if (!mp->pixelmode) {
				//				XSetForeground(display, gc, MI_BLACK_PIXEL(mi));
				//				XDrawLines(display, window, gc, p, 5, CoordModeOrigin);
				poly(screen, p, 4, Endsquare, Endsquare, 1, display->black, screen->r.min);
			}
		}
	}
	mp->x++;
	if (mp->x == WORLDWIDTH - 1) {
		mp->y++;
		mp->x = 0;
	}
	if (mp->y == WORLDWIDTH - 1) {
		mp->stage++;
		flushimage(display, 1);
	}
}

void
init_mountain(void)
{
	int         i, j, x, y;
	mountainstruct *mp;

	if (mountains == NULL) {
		if ((mountains = (mountainstruct *) calloc(1,
			sizeof (mountainstruct))) == NULL)
			return;
	}
	mp = &mountains[0];

	mp->width = Dx(screen->r);
	mp->height = Dy(screen->r);
	mp->pixelmode = (mp->width + mp->height < 200);
	mp->stage = 0;
	mp->time = 0;
	mp->x = mp->y = 0;
	mp->joke = (Bool) (NRAND(10) == 0);
	mp->wireframe = (Bool) (LRAND() & 1);

	draw(screen, screen->r, display->black, nil, ZP);
	for (y = 0; y < (int) WORLDWIDTH; y++)
		for (x = 0; x < (int) WORLDWIDTH; x++)
			mp->h[x][y] = 0;

	j = COUNT;
	if (j < 0)
		j = NRAND(-j) + 1;
	for (i = 0; i < j; i++)
		mp->h[RANGE_RAND(1, WORLDWIDTH - 1)][RANGE_RAND(1, WORLDWIDTH - 1)] =
			NRAND(MAXHEIGHT);

	for (y = 0; y < WORLDWIDTH; y++)
		for (x = 0; x < WORLDWIDTH; x++)
			spread(mp->h, x, y);

	for (y = 0; y < WORLDWIDTH; y++)
		for (x = 0; x < WORLDWIDTH; x++) {
			mp->h[x][y] = mp->h[x][y] + NRAND(10) - 5;
			if (mp->h[x][y] < 10)
				mp->h[x][y] = 0;
		}

	mp->offset = NRAND(255) + 1;
}

void
draw_mountain(void)
{
	mountainstruct *mp;

	if (mountains == NULL)
		return;
	mp = &mountains[0];


	switch (mp->stage) {
	case 0:
		drawamountain();
		break;
	case 1:
		if (++mp->time > CYCLES)
			mp->stage++;
		break;
	case 2:
		init_mountain();
		break;
	}
	flushimage(display,1);
	frames++;
}

void
release_mountain(void)
{
	if (mountains != NULL) {

		free((void *) mountains);
		mountains = (mountainstruct *) NULL;
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(mountains != nil)
		release_mountain();
	init_mountain();
}

void 
threadmain(int argc, char **argv)
{
	Mouse m;
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "mountain") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(256)));
	}

	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_mountain();
	}
}
