/* munch.c
 * A munching squares implementation for X
 * Tim Showalter <tjs@andrew.cmu.edu>
 * 
 * Copyright 1997, Tim Showalter
 * Permission is granted to copy, modify, and use this as long
 * as this notice remains intact.  No warranties are expressed or implied.
 * CMU Sucks.
 * 
 * Some code stolen from / This is meant to work with
 * xscreensaver, Copyright (c) 1992, 1995, 1996
 * Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* Munching Squares is this simplistic, silly screen hack (according
   to HAKMEM, discovered by Jackson Wright in 1962) where you take
   Y = X XOR T and graph it over and over.  According to HAKMEM, it 
   takes 5 instructions of PDP-1 assembly.  This is a little more
   complicated than that, mostly X's fault, but it does some other
   random things.

   http://www.inwap.com/pdp10/hbaker/hakmem/hacks.html#item146
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 08/02
  */

/* plan9-related stuff */
#include "plan9.h"
static int col;


/* flags for random things.  Must be < log2(random's maximum), incidentially.
 */
#define SHIFT_KX (0x01)
#define SHIFT_KT (0x02)
#define SHIFT_KY (0x04)
#define GRAV     (0x08)





/* only configurable things right now */
static int delay, hold, clear, logminwidth, shiftk, xor;

static void munchOnce (int width, /* pixels */
int atX, int atY, /* pixels */
int kX, int kT, int kY, /* pixels */
int grav /* 0 or not */) {

	int x, y, t;
	int drawX, drawY;
	Point p = screen->r.min;
	Rectangle r;

	/* Finally draw this munching square. */
	for(t = 0; t < width; t++) {
		for(x = 0; x < width; x++) {
			/* figure out the next point */
			y = ((x ^ ((t + kT) % width)) + kY) % width;

			drawX = ((x + kX) % width) + atX;
			drawY = (grav ? y + atY : atY + width - 1 - y);


			//	    XDrawPoint(dpy, w, gc, drawX, drawY);
			p = addpt(screen->r.min, Pt(drawX, drawY));
			r = Rpt(p, addpt(p, Pt(1, 1)));
			draw(screen, r, colors[col], nil, ZP);
		}

		/* we've drawn one pass' worth of points.  let the server catch up
	   or this won't be interesting to watch at all.  we call this here
	   in the hopes of having one whole sequence of dots appear at the
	   same time (one for each value of x, surprisingly enough)
	   */
		flushimage(display, 1);
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
			}
		}
		sleep(5);
	}
}

/*
 * dumb way to get # of digits in number.  Probably faster than actually
 * doing a log and a division, maybe.
 */
static int dumb_log_2(int k) {
	int r = -1;
	while (k > 0) {
		k >>= 1;
		r++;
	}
	return r;
}

/* This parses arguments, initializes the window, etc., and finally starts
 * calling munchOnce ad infinitum.
 */
void
screenhack (void)
{
	int logmaxwidth;
	int n = 0;  /* number of squares before we have to clear */
	int randflags;
	int thiswidth;


	/* We need a square; limit on screen size? */
	/* we want a power of 2 for the width or the munch doesn't fill up.
     */
	logmaxwidth = (int)
		dumb_log_2(Dx(screen->r) < Dy(screen->r) ? Dx(screen->r) : Dy(screen->r));


	if (logmaxwidth < logminwidth) {
		/* off-by-one error here?  Anyone running on < 640x480? */
		fprint(2, "munch: screen too small; use -logminwidth\n");
		fprint(2, "\t(width is %d; log is %d; log must be at least "
			"%d)\n", 
			(Dx(screen->r) < Dy(screen->r) ? Dx(screen->r) : Dy(screen->r)),
			logmaxwidth, logminwidth);
		threadexitsall("screen too small");
	}

	clear = 50;
	logminwidth = 7;
	shiftk = 1;
	xor = 1;

	draw(screen, screen->r, display->black, nil, ZP);
	for(;;) {
		/* saves some calls to random.  big deal */
		randflags = random();

		/* choose size -- power of two */
		thiswidth = 1 << (logminwidth +
			(random() % (1 + logmaxwidth - logminwidth)));
		col = nrand(255);
		munchOnce(
			thiswidth, /* Width, in pixels */

		/* draw at this location */
		(random() % (Dx(screen->r) <= thiswidth ? 1
			: Dx(screen->r) - thiswidth)),
			(random() % (Dy(screen->r) <= thiswidth ? 1
			: Dx(screen->r) - thiswidth)),

			/* wrap-around by these values; no need to %
		     as we end up doing that later anyway*/
		((shiftk && (randflags & SHIFT_KX))
			? (random() % thiswidth) : 0),
			((shiftk && (randflags & SHIFT_KT))
			? (random() % thiswidth) : 0),
			((shiftk && (randflags & SHIFT_KY))
			? (random() % thiswidth) : 0),

			/* set the gravity of the munch, or rather,
		     which direction we draw stuff in. */
		(randflags & GRAV)
			);

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
			}
		}
		frames++;

		if (clear && ++n >= clear) {
			draw(screen, screen->r, display->black, nil, ZP);
			n = 0;
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}
	sysfatal("cannot resize");
}


void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "munch") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++)
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), RGBA32, 1, setalpha(cmap2rgba(nrand(65535)), nrand(0xff)));

	einit(Emouse);
	screenhack();
}
