/* pacman --- Mr. Pacman and his ghost friends */

/*-
 * Copyright (c) 2002 by Edwin de Jong <mauddib@gmx.net>.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 13-May-2002: Added -trackmouse feature thanks to code from 'maze.c'.  
 *		splitted up code into several files.  Retouched AI code, cleaned
 *		up code.
 *  3-May-2002: Added AI to pacman and ghosts, slowed down ghosts.
 * 26-Nov-2001: Random level generator added
 * 01-Nov-2000: Allocation checks
 * 04-Jun-1997: Compatible with xscreensaver
 *
 */

/* TODO:
   1. add "bonus" dots
   2. make better ghost sprites (eyes, waving dress)
   3. make a bit better pacman sprite (mouth should be larger)
   4. think of a better level generation algorithm
*/


/*
  * ported to Plan 9 by andrey@lanl.gov, 07/02
  */

/* plan9-related stuff */
#include "plan9.h"

Mouse m;


#include "pacman.h"
#include "pacman_ai.h"
#include "pacman_level.h"

#define None nil

enum {
	Blue = 0,
	Yellow,
	Green,
};


Image * 	currentcolor;

static void
free_pacman(pacmangamestruct *pp)
{
	int         dir, mouth;

	if (pp->ghosts != NULL) {
		free(pp->ghosts);
		pp->ghosts = (ghoststruct *) NULL;
	}
	if (pp->ghostPixmap != None) {
		freeimage(pp->ghostPixmap);
		pp->ghostPixmap = None;
	}
	for (dir = 0; dir < 4; dir++)
		for (mouth = 0; mouth < MAXMOUTH; mouth++)
			if (pp->pacmanPixmap[dir][mouth] != None) {
				freeimage(pp->pacmanPixmap[dir][mouth]);
				pp->pacmanPixmap[dir][mouth] = None;
			}
}

/* Checks for death of any ghosts/pacman and updates.  It also makes a new
   level if all ghosts are dead or all dots are eaten. */
static void
check_death(pacmangamestruct *pp)
{
	unsigned int	ghost;
	int	alldead;
	Point p; 
	Rectangle r;

	alldead = 1;
	for (ghost = 0; ghost < pp->nghosts; ghost++) {
		if (pp->ghosts[ghost].dead == True)
			continue;

		if ((pp->ghosts[ghost].nextrow == NOWHERE &&
			pp->ghosts[ghost].nextcol == NOWHERE)) {
			alldead = 0;
			continue;
		}

		if (((pp->ghosts[ghost].nextrow == pp->pacman.nextrow) &&
			(pp->ghosts[ghost].nextcol == pp->pacman.nextcol)) ||
			((pp->ghosts[ghost].nextrow == pp->pacman.row) &&
			(pp->ghosts[ghost].nextcol == pp->pacman.col) &&
			(pp->ghosts[ghost].row == pp->pacman.nextrow) &&
			(pp->ghosts[ghost].col == pp->pacman.nextcol))) {
			pp->ghosts[ghost].dead = 1;
			//XSetForeground(display, 
			//		     pp->stippledGC, 
			//		     MI_BLACK_PIXEL(mi));
			//XFillRectangle(display, window, 
			//		    pp->stippledGC,
			//		    pp->ghosts[ghost].cf, 
			//		    pp->ghosts[ghost].rf, 
			//		    pp->spritexs, 
			//		    pp->spriteys);
			p = Pt(pp->ghosts[ghost].cf, pp->ghosts[ghost].rf);
			r = Rpt(p, addpt(p, Pt(pp->spritexs, pp->spriteys)));
			draw(img, r, display->black, nil, ZP);
		} else
			alldead = 0;
	}

	if (alldead == 1 || pp->dotsleft == 0)
		repopulate();
}

/* Resets state of ghosts + pacman.  Creates a new level, draws that level. */
static void
repopulate(void)
{
	pacmangamestruct *pp = &pacmangames;
	unsigned int ghost;
	int i = createnewlevel();

	draw(img, img->r, display->black, nil, ZP);
	drawlevel();

	pp->gamestate = GHOST_DANGER;

	pp->pacman.row = (LEVHEIGHT + JAILHEIGHT)/2 - i;
	pp->pacman.col = (LEVWIDTH/2);
	pp->pacman.nextrow = NOWHERE;
	pp->pacman.nextcol = NOWHERE;
	pp->pacman.cf = NOWHERE;
	pp->pacman.rf = NOWHERE;
	pp->pacman.oldcf = NOWHERE;
	pp->pacman.oldrf = NOWHERE;
	pp->pacman.oldlx = NOWHERE;
	pp->pacman.oldly = NOWHERE;
	pp->pacman.aistate = ps_eating;
	pp->pacman.cur_trace = 0;
	pp->pacman.roundscore = 0;
	pp->pacman.speed = 4;
	pp->pacman.lastturn = 0;
	pp->pacman.delta.x = 0;
	pp->pacman.delta.y = 0;
	pac_clear_trace(&(pp->pacman));

	for (ghost = 0; ghost < pp->nghosts; ghost++) {
		pp->ghosts[ghost].col = (LEVWIDTH/2);
		pp->ghosts[ghost].row = (LEVHEIGHT/2);
		pp->ghosts[ghost].nextcol = NOWHERE;
		pp->ghosts[ghost].nextrow = NOWHERE;
		pp->ghosts[ghost].dead = 0;
		pp->ghosts[ghost].lastbox = START;
		pp->ghosts[ghost].cf = NOWHERE;
		pp->ghosts[ghost].rf = NOWHERE;
		pp->ghosts[ghost].oldcf = NOWHERE;
		pp->ghosts[ghost].oldrf = NOWHERE;
		pp->ghosts[ghost].aistate = inbox;
		pp->ghosts[ghost].timeleft = ghost * 50;
		pp->ghosts[ghost].speed = 3;
		pp->ghosts[ghost].delta.x = 0;
		pp->ghosts[ghost].delta.y = 0;

		ghost_update(pp, &(pp->ghosts[ghost]));
	}
	check_death(pp);
	pac_update(pp, &(pp->pacman));
}

/* Sets the color to the color of a wall. */
static void 
setwallcolor(void)
{
	currentcolor = colors[Blue];
}

/* Sets the color to the color of a dot. */
static void 
setdotcolor(void)
{
	currentcolor = display->white;
}

/* Draws a block in the level at the specified x and y locations. */
static void
drawlevelblock(pacmangamestruct *pp, 
const unsigned x, const unsigned y)
{
	int dx = 0, dy = 0;
	Point p;
	Rectangle r;

	if (pp->xs % 2 == 1) dx = -1;
	if (pp->ys % 2 == 1) dy = -1;

	if (pp->xs < 2 || pp->ys < 2) {
		switch(pp->level[y*LEVWIDTH + x]) {
		case ' ':
		case '=':
			break;
		case '.':
			setdotcolor();
			//(void)XDrawPoint(display, window,
			//		 pp->stippledGC,
			//		 x * pp->xs + pp->xb, 
			//		 y * pp->ys + pp->yb);
			p = Pt(x * pp->xs + pp->xb,y * pp->ys + pp->yb);
			r = Rpt(p, addpt(p, Pt(1, 1)));
			draw(img, r, currentcolor, nil, ZP);
			break;
		default:
			setwallcolor();
			//(void)XDrawPoint(display, window,
			//		 pp->stippledGC,
			//		 x * pp->xs + pp->xb, 
			//		 y * pp->ys + pp->yb);
			p = Pt(x * pp->xs + pp->xb, y * pp->ys + pp->yb);
			r = Rpt(p, addpt(p, Pt(1, 1)));
			draw(img, r, currentcolor, nil, ZP);
		}

		return;
	}

	switch (pp->level[y*LEVWIDTH + x]) {
	case ' ':
	case '=':
		break;

	case '.':
		setdotcolor();
		if (pp->xs < 8 || pp->ys < 8) {
			//(void)XDrawPoint(display, window,
			//		 pp->stippledGC,
			//		 x * pp->xs + pp->xb +
			//		 pp->xs/2, 
			//		 y * pp->ys + pp->yb +
			//		 pp->ys/2);
			p = Pt(x * pp->xs + pp->xb +
				pp->xs/2,  y * pp->ys + pp->yb +
				pp->ys/2);
			r = Rpt(p, addpt(p, Pt(1, 1)));
			draw(img, r, currentcolor, nil, ZP);
			break;
		}

		ellipse(img, 
			Pt((pp->xs * x) + 
			(pp->xs / 2) - 
			(pp->xs > 32 ? (pp->xs / 16) : 1) +
			pp->xb,
			(pp->ys * y) + 
			(pp->ys / 2) -
			(pp->ys > 32 ? (pp->ys / 16) : 1) +
			pp->yb), 
			(pp->xs > 32 ? (pp->xs / 32) : 1),
			(pp->ys > 32 ? (pp->ys / 32) : 1), 
			0, currentcolor, ZP);
		break;

	case '-':
		setwallcolor();
		line(img, Pt((pp->xs * x) + pp->xb,
			(pp->ys * y) + (pp->ys / 2) + pp->yb), 
			Pt((pp->xs * (x + 1)) + pp->xb,
			(pp->ys * y) + (pp->ys / 2) + pp->yb), 
			Enddisc, Enddisc, 0, currentcolor, ZP);
		break;

	case '|':
		setwallcolor();
		line(img, Pt((pp->xs * x) + (pp->xs / 2) + pp->xb,
			(pp->ys * y) + pp->yb), 
			Pt((pp->xs * x) + (pp->xs / 2) + pp->xb,
			(pp->ys * (y + 1)) + pp->yb), 
			Enddisc, Enddisc, 0, currentcolor, ZP);
		break;

	case '_':
		setwallcolor();
		arc(img, 
			Pt((pp->xs * x) - (pp->ys / 2) + pp->xb + dx + pp->xs/2,
			(pp->ys * y) + (pp->ys / 2) + pp->yb + pp->ys/2), 
			pp->xs/2, pp->ys/2,
			0, currentcolor, ZP, 0, 90);
		break;

	case ',':
		setwallcolor();
		arc(img, 
			Pt((pp->xs * x) + (pp->ys / 2) + pp->xb + pp->xs/2,
			(pp->ys * y) + (pp->ys / 2) + pp->yb + pp->ys/2), 
			pp->xs/2, pp->ys/2,
			0, currentcolor, ZP, 90, 90);
		break;

	case '`':
		setwallcolor();
		arc(img, 
			Pt((pp->xs * x) + (pp->ys / 2) + pp->xb+pp->xs/2,
			(pp->ys * y) - (pp->ys / 2) + pp->yb + dy+pp->ys/2), 
			pp->xs/2, pp->ys/2,
			0, currentcolor, ZP, 180, 90);
		break;

	case '\'':
		setwallcolor();
		arc(img, 
			Pt((pp->xs * x) - (pp->ys / 2) + pp->xb + dx + pp->xs/2,
			(pp->ys * y) - (pp->ys / 2) + pp->yb + dy + pp->ys/2), 
			pp->xs/2, pp->ys/2,
			0, currentcolor, ZP, 270, 90);
		break;

	}
}

/* Draws a complete level. */
static void
drawlevel(void)
{
	pacmangamestruct *pp = &pacmangames;
	unsigned int x, y;

	for (y = 0; y < LEVHEIGHT; y++)
		for (x = 0; x < LEVWIDTH; x++)
			drawlevelblock(pp, x, y);
}


/* Draws the pacman sprite, removing the previous location. */
static void
draw_pacman_sprite(void)
{
	pacmangamestruct *pp = &pacmangames;
	unsigned int dir;

	pp->pacman.cf = pp->pacman.col * pp->xs + pp->pacman.delta.x * 
		pp->pacman.cfactor + pp->xb + pp->spritedx;
	pp->pacman.rf = pp->pacman.row * pp->ys + pp->pacman.delta.y * 
		pp->pacman.rfactor + pp->yb + pp->spritedy;

	dir = (ABS(pp->pacman.cfactor) * (2 - pp->pacman.cfactor) +
		ABS(pp->pacman.rfactor) * (1 + pp->pacman.rfactor)) % 4;

	if (pp->pacman.oldcf != NOWHERE && pp->pacman.oldrf != NOWHERE) {
		//XFillRectangle(display, window, pp->stippledGC,
		//	     pp->pacman.oldcf, pp->pacman.oldrf, 
		//	     pp->spritexs, pp->spriteys);
		draw(img, Rect(pp->pacman.oldcf, pp->pacman.oldrf, 
			pp->pacman.oldcf + pp->spritexs, pp->pacman.oldrf + pp->spriteys),
			display->black, nil, ZP);
	}

	currentcolor = colors[Yellow];
	draw(img, Rect(pp->pacman.cf, pp->pacman.rf, 
		pp->pacman.cf+pp->spritexs, pp->pacman.rf + pp->spriteys),
		currentcolor, pp->pacmanPixmap[dir][pp->pacman.mouthstage], ZP);


	pp->pacman.mouthstage += pp->pacman.mouthdirection;
	if ((pp->pacman.mouthstage >= MAXMOUTH) ||
		(pp->pacman.mouthstage < 0)) {
		pp->pacman.mouthdirection *= -1;
		pp->pacman.mouthstage += pp->pacman.mouthdirection * 2;
	}
	pp->pacman.oldcf = pp->pacman.cf;
	pp->pacman.oldrf = pp->pacman.rf;
}

/* Draws a ghost sprite, removing the previous sprite and restores the level. */
static void
draw_ghost_sprite(const unsigned ghost) {
	pacmangamestruct *pp = &pacmangames;

	pp->ghosts[ghost].cf = 
		pp->ghosts[ghost].col * pp->xs + pp->ghosts[ghost].delta.x * 
		pp->ghosts[ghost].cfactor + pp->xb + pp->spritedx;
	pp->ghosts[ghost].rf = 
		pp->ghosts[ghost].row * pp->ys + pp->ghosts[ghost].delta.y * 
		pp->ghosts[ghost].rfactor + pp->yb + pp->spritedy;

	if (pp->ghosts[ghost].oldcf != NOWHERE ||
		pp->ghosts[ghost].oldrf != NOWHERE) {
		draw(img, Rect(pp->ghosts[ghost].oldcf, pp->ghosts[ghost].oldrf, 
			pp->ghosts[ghost].oldcf+pp->spritexs, pp->ghosts[ghost].oldrf + pp->spriteys),
			display->black, nil, ZP);

	}


	drawlevelblock(pp, 
		(unsigned int)pp->ghosts[ghost].col, 
		(unsigned int)pp->ghosts[ghost].row);

	currentcolor = colors[Green];
	draw(img, Rect(pp->ghosts[ghost].cf, pp->ghosts[ghost].rf, 
		pp->ghosts[ghost].cf+pp->spritexs, 
		pp->ghosts[ghost].rf + pp->spriteys),
		currentcolor, pp->ghostPixmap, ZP);

	pp->ghosts[ghost].oldcf = pp->ghosts[ghost].cf;
	pp->ghosts[ghost].oldrf = pp->ghosts[ghost].rf;
}

/* Does all drawing of moving sprites in the level. */
static void
pacman_tick(void)
{
	pacmangamestruct *pp = &pacmangames;
	unsigned int ghost;

	for (ghost = 0; ghost < pp->nghosts; ghost++) {
		if (pp->ghosts[ghost].dead == True)
			continue;
		draw_ghost_sprite(ghost);
	}

	draw_pacman_sprite();

	draw(screen, screen->r, img, nil, ZP);

	flushimage(display, 1);
}

/* Hook function, sets state to initial position. */
void
init_pacman(void)
{
	pacmangamestruct *pp;
	int	    dir, mouth;
	XPoint	    points[9];
	int 		size = 0;

	pp = &pacmangames;

	pp->width = Dx(screen->r);
	pp->height = Dy(screen->r);
	if (pp->ghostPixmap != None) {
		freeimage(pp->ghostPixmap);
		pp->ghostPixmap = None;
		pp->graphics_format = 0 /*IS_NONE*/;
	}

	if (size == 0 ||
		MINGRIDSIZE * size > (int)pp->width ||
		MINGRIDSIZE * size > (int)pp->height) {

		pp->ys = pp->xs = MAX(MIN(pp->width/LEVWIDTH, 
			pp->height/LEVHEIGHT), 1);
	} else {
		if (size < -MINSIZE)
			pp->ys = (short)(NRAND( MIN( -size, MAX( MINSIZE, 
				MIN( pp->width, pp->height) / MINGRIDSIZE))
				- MINSIZE + 1) + MINSIZE);
		else if (size < MINSIZE)
			pp->ys = MINSIZE;
		else
			pp->ys = (short)(MIN(size, 
				MAX(MINSIZE, MIN(pp->width, pp->height) /
				MINGRIDSIZE)));
		pp->xs = pp->ys;
	}

	pp->wallwidth = (unsigned int)(pp->xs + pp->ys) >> 4;
	if (pp->wallwidth < 1) pp->wallwidth = 1;
	pp->incx = (pp->xs >> 3) + 1;
	pp->incy = (pp->ys >> 3) + 1;
	pp->ncols = (unsigned short)MAX(LEVWIDTH, 2);
	pp->nrows = (unsigned short)MAX(LEVHEIGHT, 2);
	pp->xb = (pp->width - pp->ncols * pp->xs) >> 1;
	pp->yb = (pp->height - pp->nrows * pp->ys) >> 1;
	pp->spritexs = MAX(pp->xs + (pp->xs >> 1) - 1, 1);
	pp->spriteys = MAX(pp->ys + (pp->ys >> 1) - 1, 1);
	pp->spritedx = (pp->xs - pp->spritexs) >> 1;
	pp->spritedy = (pp->ys - pp->spriteys) >> 1;


#define SETPOINT(p, xp, yp) p.x = xp; p.y = yp

	/* draw the triangles on the bottom (scalable) */
	SETPOINT(points[0], 0, 				pp->spriteys * 5 / 6);
	SETPOINT(points[1], pp->spritexs / 6,		pp->spriteys);
	SETPOINT(points[2], pp->spritexs / 3,		pp->spriteys * 5 / 6);
	SETPOINT(points[3], pp->spritexs / 2,		pp->spriteys);
	SETPOINT(points[4], pp->spritexs * 2 / 3,	pp->spriteys * 5 / 6);
	SETPOINT(points[5], pp->spritexs * 5 / 6,	pp->spriteys);
	SETPOINT(points[6], pp->spritexs,		pp->spriteys * 5 / 6);
	SETPOINT(points[7], pp->spritexs,		pp->spriteys / 2);
	SETPOINT(points[8], 0, 				pp->spriteys / 2);


	pp->ghostPixmap = allocimage(display, Rect(0, 0, pp->spritexs, pp->spriteys), screen->chan, 0, DBlack);
	if(pp->ghostPixmap == nil)
		sysfatal("cannot allocate ghost pixmap image");

	fillarc(pp->ghostPixmap, Pt(pp->spritexs/2, pp->spriteys/2), pp->spritexs/2, pp->spriteys/2,  display->white, ZP, 0, 180);
	fillpoly(pp->ghostPixmap, points, 9, 0, display->white, ZP);

	for (dir = 0; dir < 4; dir++)
		for (mouth = 0; mouth < MAXMOUTH; mouth++) {
			pp->pacmanPixmap[dir][mouth] = 
				allocimage(display, Rect(0, 0, pp->spritexs, pp->spriteys), 
				screen->chan, 0, DBlack);

			if(pp->pacmanPixmap[dir][mouth] == nil)
				sysfatal("cannot allocate one of the pacman pixmap images");

			if (pp->spritexs == 1 && pp->spriteys == 1)
				draw(pp->pacmanPixmap[dir][mouth], 
					Rect(0, 0, pp->spritexs, pp->spriteys), display->white, nil, ZP);
			else
				fillarc(pp->pacmanPixmap[dir][mouth], 
					Pt(pp->spritexs/2, pp->spriteys/2), 
					pp->spritexs/2, pp->spriteys/2, 
					display->white, ZP, 
					((90 - dir * 90) + mouth * 5),  (360 + (-2 * mouth * 5)));

		}

	pp->pacman.lastbox = START;
	pp->pacman.mouthdirection = 1;
	pp->pacman.nextcol = NOWHERE;
	pp->pacman.nextrow = NOWHERE;

	if (pp->ghosts != NULL) {
		free(pp->ghosts);
		pp->ghosts = (ghoststruct *) NULL;
	}
	pp->nghosts = GHOSTS;

	if (!pp->ghosts)
		if ((pp->ghosts = (ghoststruct *) calloc((uint)pp->nghosts,
			sizeof (ghoststruct))) == NULL) {
			sysfatal("ghosts calloc: %r");
			return;
		}

	pp->pacman.mouthstage = MAXMOUTH - 1;

	draw(img, img->r, display->black, nil, ZP);
	draw(screen, screen->r, display->black, nil, ZP);
	repopulate();
}

/* Callback function called for each tick.  This is the complete machinery of
   everything that moves. */
void
draw_pacman(void)
{
	unsigned int g;
	pacmangamestruct *pp;

	pp = &pacmangames;
	if (pp->ghosts == NULL)
		return;

	pp->pacman.err.x = (pp->pacman.err.x + 1) % pp->pacman.speed;
	pp->pacman.err.y = (pp->pacman.err.y + 1) % pp->pacman.speed;
	pp->pacman.delta.x += pp->pacman.err.x != 0 ? pp->incx : 0;
	pp->pacman.delta.y += pp->pacman.err.y != 0 ? pp->incy : 0;

	if (pp->pacman.delta.x >= pp->xs && pp->pacman.delta.y >= pp->ys) {
		pac_update(pp, &(pp->pacman));
		check_death(pp);
		pp->pacman.delta.x = pp->incx;
		pp->pacman.delta.y = pp->incy;
	}

	if (pp->pacman.delta.x > pp->xs + pp->incx)
		pp->pacman.delta.x = pp->xs + pp->incx;
	if (pp->pacman.delta.y > pp->ys + pp->incy)
		pp->pacman.delta.y = pp->ys + pp->incy;

	for (g = 0; g < pp->nghosts; g++) {
		if (pp->ghosts[g].dead == True) continue;

		pp->ghosts[g].err.x = (pp->ghosts[g].err.x + 1) % 
			pp->ghosts[g].speed;
		pp->ghosts[g].err.y = (pp->ghosts[g].err.y + 1) % 
			pp->ghosts[g].speed;
		pp->ghosts[g].delta.x += pp->ghosts[g].err.x != 0 ? 
			pp->incx : 0;
		pp->ghosts[g].delta.y += pp->ghosts[g].err.y != 0 ? 
			pp->incy : 0;

		if (pp->ghosts[g].delta.x >= pp->xs && 
			pp->ghosts[g].delta.y >= pp->ys) {
			ghost_update(pp, &(pp->ghosts[g]));
			pp->ghosts[g].delta.x = pp->incx;
			pp->ghosts[g].delta.y = pp->incy;
		}

		if (pp->ghosts[g].delta.x > pp->xs + pp->incx)
			pp->ghosts[g].delta.x = pp->xs + pp->incx;
		if (pp->ghosts[g].delta.y > pp->ys + pp->incy)
			pp->ghosts[g].delta.y = pp->ys + pp->incy;
	}

	pacman_tick();
}

/* Refresh current level. */
void
refresh_pacman(void)
{
	drawlevel();
	pacman_tick();
}

/* Callback to change level. */
void
change_pacman(void)
{
	draw(screen, screen->r, display->black, nil, ZP);
	repopulate( );
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}


	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DBlack);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");
	free_pacman(&pacmangames);
	init_pacman();

}

void 
threadmain(int argc, char **argv)
{
	Mouse m;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	case 't':	/* track mouse */
		trackmouse = 1;
		break;
	}ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "pacman") < 0)
		sysfatal("initdraw failed: %r");

	colors[Blue] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, DPaleblue);
	colors[Yellow] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, DPaleyellow);
	colors[Green] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, DPalegreen);


	if(colors[Blue] == nil || colors[Yellow] == nil || colors[Green] == nil)
		sysfatal("cannot allocate colors: %r");

	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_pacman();
		frames++;
	}
}
