/* Copyright (c) 2003 Levi Burton <donburton@sbcglobal.net>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"

typedef struct _square {
	int x, y, w, h;
	int color;
} square;

void
randomize_square_colors(square *squares, int nsquares, int ncolors)
{
	int i;
	square *s = squares;
	for (i = 0; i < nsquares; i++)
		s[i].color = random() % ncolors;
}

int sw, sh, gw, gh, nsquares;

square *squares = nil;
int delay = 25;
int subdivision = 10;
int ncolors = 256;
Bool twitch;

void
init_squares(void)
{
	int x, y;

	twitch = nrand(2);
	subdivision = nrand(10) + 5;

	sw = Dx(screen->r) / subdivision;
	sh = Dy(screen->r) / subdivision;
	gw = Dx(screen->r) / sw;
	gh = Dy(screen->r) / sh;
	nsquares = gw * gh;


	if(squares != nil)
		free(squares);

	squares = (square *) calloc (nsquares, sizeof(square));

	for (y = 0; y < gh; y++)
		for (x = 0; x < gw; x++)
		{
			square *s = (square *) &squares[gw * y + x];
			s->w = sw;
			s->h = sh;
			s->x = x * sw;
			s->y = y * sh;
		}

	randomize_square_colors(squares, nsquares, ncolors);

}

void 
screenhack (void)
{

	int x, y;


	while (1)
	{
		for (y = 0; y < gh; y++)
			for (x = 0; x < gw; x++)
			{
				square *s = (square *) &squares[gw * y + x];
				//XSetForeground (dpy, gc, colors[s->color].pixel);
				//XFillRectangle (dpy, b, gc, s->x, s->y, 
				//    border ? s->w - border : s->w, 
				//    border ? s->h - border : s->h);

				draw(screen, rectaddpt(
					Rect(s->x, s->y,s->x+s->w - 1,s->y+s->h - 1),screen->r.min), 
					colors[s->color], nil, ZP);

				s->color++;
				if (s->color == ncolors)
				{
					if (twitch && ((random() % 4) == 0))
						randomize_square_colors (squares, nsquares, ncolors);
					else
						s->color = random() % ncolors;
				}
			}

		flushimage(display, 1);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if (delay)
			sleep (delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}


	init_squares();

	draw(screen, screen->r, display->black, nil, ZP);

}

void 
threadmain(int argc, char **argv)
{
	int i;
	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "popsquares") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			i<<8 | i<<12 | 0xff);
	}

	einit(Emouse);
	eresized(0);

	screenhack();
}
