/* xscreensaver, Copyright (c) 1992, 1995, 1997, 1998, 1999
 *  Jamie Zawinski <jwz@jwz.org>
 *
 *  reaction/diffusion textures
 *  Copyright (c) 1997 Scott Draves spot@transmeta.com
 *  this code is derived from Bomb
 *  see http://www.cs.cmu.edu/~spot/bomb.html
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * And remember: X Windows is to graphics hacking as roman numerals are to
 * the square root of pi.
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 08/02
  */

/* plan9-related stuff */
#include "plan9.h"

#include "smooth.h"


uchar col[256];

Image *mybitmap;


/* costs ~6% speed */
#define dither_when_mapped 1

static int verbose;
static int ncolors = 0;
static unsigned char *mc = 0;
static int mapped;
static int pdepth;
static void random_colors(void);

/* -----------------------------------------------------------
   pixel hack, 8-bit pixel grid, first/next frame interface

   pixack_init(int *size_h, int *size_v)
   pixack_frame(char *pix_buf)
   */


#define bps 16
#define mx ((1<<16)-1)

/* you can replace integer mults wish shift/adds with these,
   but it doesn't help on my 586 */
#define x5(n) ((n<<2)+n)
#define x7(n) ((n<<3)-n)

/* why strip bit? */
#define R (random()&((1<<30)-1))

static int frame = 0, epoch_time;
static ushort *r1, *r2, *r1b, *r2b;
static int width, height, npix;
static int radius;
static int reaction = 0;
static int diffusion = 0;

/* returns number of pixels that the pixack produces.  called once. */
static void
pixack_init(int *size_h, int *size_v)
{
	int sz_base;
	width =-1;
	height = -1;
	sz_base = 80 + (R%40);
	if (width <= 0) width = (R%20) ? sz_base : (28 + R%10);
	if (height <= 0) height = (R%20) ? sz_base : (28 + R%10);

	/* jwz: when (and only when) XSHM is in use on an SGI 8-bit visual,
     we get shear unless width is a multiple of 4.  I don't understand
     why.  This is undoubtedly the wrong fix... */
	width &= ~0x7;

	/* don't go there */
	if (width < 10) width = 10;
	if (height < 10) height = 10;
	epoch_time = 40000;
	npix = (width + 2) * (height + 2);
	r1 = (ushort *) malloc(sizeof(ushort) * npix);
	r2 = (ushort *) malloc(sizeof(ushort) * npix);
	r1b = (ushort *) malloc(sizeof(ushort) * npix);
	r2b = (ushort *) malloc(sizeof(ushort) * npix);

	if (!r1 || !r2 || !r1b || !r2b) {
		fprint(2, "not enough memory for %d pixels.\n", npix);
		threadexitsall("malloc");
	}

	*size_h = width;
	*size_v = height;
}

#define test_pattern_hyper 0


/* returns the pixels.  called many times. */
static void
pixack_frame(char *pix_buf)
{
	int i, j;
	int w2 = width + 2;
	ushort *t;

	if (!(frame%epoch_time)) {
		int s;
		if (0 != frame) {
			int t = epoch_time / 500;
			if (t > 15)
				t = 15;
			sleep(t);
		}

		for (i = 0; i < npix; i++) {
			/* equilibrium */
			r1[i] = 65500;
			r2[i] = 11;
		}

		//    XSetWindowBackground(display, window, col[255 % ncolors].pixel);
		//    XClearWindow(display, window);
		draw(screen, screen->r, display->black, nil, ZP);

		s = w2 * (height/2) + width/2;
		radius = -1;
		{
			int maxr = width/2-2;
			int maxr2 = height/2-2;
			if (maxr2 < maxr) maxr = maxr2;

			if (radius < 0)
				radius = 1 + ((R%10) ? (R%5) : (R % maxr));
			if (radius > maxr) radius = maxr;
		}
		for (i = -radius; i < (radius+1); i++)
			for (j = -radius; j < (radius+1); j++)
				r2[s + i + j*w2] = mx - (R&63);
		reaction = -1;
		if (reaction < 0 || reaction > 2) reaction = R&1;
		diffusion = -1;
		if (diffusion < 0 || diffusion > 2)
			diffusion = (R%5) ? ((R%3)?0:1) :
			2;
		if (2 == reaction && 2 == diffusion)
			reaction = diffusion = 0;

	}
	for (i = 0; i <= width+1; i++) {
		r1[i] = r1[i + w2 * height];
		r2[i] = r2[i + w2 * height];
		r1[i + w2 * (height + 1)] = r1[i + w2];
		r2[i + w2 * (height + 1)] = r2[i + w2];
	}
	for (i = 0; i <= height+1; i++) {
		r1[w2 * i] = r1[width + w2 * i];
		r2[w2 * i] = r2[width + w2 * i];
		r1[w2 * i + width + 1] = r1[w2 * i + 1];
		r2[w2 * i + width + 1] = r2[w2 * i + 1];
	}
	for (i = 0; i < height; i++) {
		int ii = i + 1;
		char *q = pix_buf + width * i;
		ushort *i1 = r1 + 1 + w2 * ii;
		ushort *i2 = r2 + 1 + w2 * ii;
		ushort *o1 = r1b + 1 + w2 * ii;
		ushort *o2 = r2b + 1 + w2 * ii;
		for (j = 0; j < width; j++) {

			int uvv, r1 = 0, r2 = 0;
			switch (diffusion) {
			case 0:
				r1 = i1[j] + i1[j+1] + i1[j-1] + i1[j+w2] + i1[j-w2];
				r1 = r1 / 5;
				r2 = (i2[j]<<3) + i2[j+1] + i2[j-1] + i2[j+w2] + i2[j-w2];
				r2 = r2 / 12;
				break;
			case 1:
				r1 = i1[j+1] + i1[j-1] + i1[j+w2] + i1[j-w2];
				r1 = r1 >> 2;
				r2 = (i2[j]<<2) + i2[j+1] + i2[j-1] + i2[j+w2] + i2[j-w2];
				r2 = r2 >> 3;
				break;
			case 2:
				r1 = (i1[j]<<1) + (i1[j+1]<<1) + (i1[j-1]<<1) + i1[j+w2] + i1[j-w2];
				r1 = r1 >> 3;
				r2 = (i2[j]<<2) + i2[j+1] + i2[j-1] + i2[j+w2] + i2[j-w2];
				r2 = r2 >> 3;
				break;
			}

			/* John E. Pearson "Complex Patterns in a Simple System"
	 Science, July 1993 */

			uvv = (((r1 * r2) >> bps) * r2) >> bps;
			switch (reaction) {  /* costs 4% */
			case 0:
				r1 += 4 * (((28 * (mx-r1)) >> 10) - uvv);
				r2 += 4 * (uvv - ((80 * r2) >> 10));
				break;
			case 1:
				r1 += 3 * (((27 * (mx-r1)) >> 10) - uvv);
				r2 += 3 * (uvv - ((80 * r2) >> 10));
				break;
			case 2:
				r1 += 2 * (((28 * (mx-r1)) >> 10) - uvv);
				r2 += 3 * (uvv - ((80 * r2) >> 10));
				break;
			}
			if (r1 > mx) r1 = mx;
			if (r2 > mx) r2 = mx;
			if (r1 < 0) r1 = 0;
			if (r2 < 0) r2 = 0;
			o1[j] = r1;
			o2[j] = r2;

			/* this is terrible.  here i want to assume ncolors = 256.
	 should lose double indirection */

			q[j] = col[(r1>>8) % ncolors];
		}
	}
	t = r1;
	r1 = r1b;
	r1b = t;
	t = r2;
	r2 = r2b;
	r2b = t;
}


/* should factor into RD-specfic and compute-every-pixel general */
void
screenhack (void)
{
	int array_width, array_height;
	double array_x = 0, array_y = 0;
	double array_dx = 0, array_dy = 0;
	char *p;
	int npix;

	double delay = 1;



	pixack_init(&width, &height);
	array_width = width  ;
	array_height = height ;
	npix = (width + 2) * (height + 2);

	ncolors = 256;

	p = malloc(npix);
	if (!p) {
		fprint(2, "not enough memory for %d pixels.\n", npix);
		threadexitsall("malloc");
	}
	mybitmap = allocimage(display,Rect(0, 0, width, height),CMAP8,1,DCyan);
	replclipr(mybitmap, 1, Rect(0, 0, Dx(screen->r), Dy(screen->r)));
	while (1) {
		Bool bump = False;

		pixack_frame(p);
		loadimage(mybitmap, Rect(0, 0, width, height), (uchar *)p, npix);

		//    	for (i = 0; i < array_width+width; i += width)
		//      	for (j = 0; j < array_height+height; j += height)
		//			draw(screen,rectaddpt(Rect(i, j,i+width, j+height),screen->r.min), mybitmap, nil, mybitmap->r.min);
		/* draw only once -- replclipr() takes care of everything */
		draw(screen,screen->r, mybitmap, nil, mybitmap->r.min);

		array_x += array_dx;
		array_y += array_dy;
		if (array_x < 0) {
			array_x = 0;
			array_dx = -array_dx;
			bump = True;
		}
		else if (array_x > (Dx(screen->r) - array_width)) {
			array_x = (Dx(screen->r) - array_width);
			array_dx = -array_dx;
			bump = True;
		}
		if (array_y < 0) {
			array_y = 0;
			array_dy = -array_dy;
			bump = True;
		}
		else if (array_y > (Dy(screen->r) - array_height)) {
			array_y = (Dy(screen->r) - array_height);
			array_dy = -array_dy;
			bump = True;
		}

		if (bump) {
			if (random() & 1) {
				double swap = array_dx;
				array_dx = array_dy;
				array_dy = swap;
			}
		}

		flushimage(display, 1);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
			}
		}
		if (delay > 0)
			sleep(delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}
	sysfatal("cannot resize");
}

enum {
	GREY = 0,
	RAND,
	TRAND,
	RED,
	GREEN,
	BLUE,
};



void 
threadmain(int argc, char **argv)
{
	int i, cols = 0;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "rd-bomb") < 0)
		sysfatal("initdraw failed: %r");

	switch(nrand(6)) {
	case 0:
		cols = GREY;
		break;
	case 1:
		cols = RAND;
		break;
	case 2:
		cols = TRAND;
		break;
	case 3:
		cols = RED;
		break;
	case 4:
		cols = GREEN;
		break;
	case 5:
		cols = BLUE;
		break;
	};

	if(cols == RAND || cols == TRAND) {
		smoothcmap(col);
	} else {
		for(i = 1; i < 256; i++) {
			switch (cols) {
			case GREY:
				col[i] = rgb2cmap(i, i, i);
				break;
			case RED:
				col[i] = rgb2cmap(i, 0, 0);
				break;
			case GREEN:
				col[i] = rgb2cmap(0, i, 0);
				break;
			case BLUE:
				col[i] = rgb2cmap(0, 0, i);
				break;
			}
		}

	}
	einit(Emouse);
	screenhack();
}
