/* slip --- lots of slipping blits */

/*-
 * Copyright (c) 1992 by Scott Draves <spot@cs.cmu.edu>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * 01-Nov-2000: Allocation checks
 * 10-May-1997: Jamie Zawinski <jwz@jwz.org> compatible with xscreensaver
 * 01-Dec-1995: Patched for VMS <joukj@hrem.stm.tudelft.nl>
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 08/02
  */

/* plan9-related stuff */
#include "plan9.h"
#include "squeeze.h"	

typedef struct {
	int         width, height;
	int         nblits_remaining;
	int         blit_width, blit_height;
	int         mode;
	int         first_time;
	int         backwards;
	short       lasthalf;
	int         stage;
	unsigned long r;
} slipstruct;
static slipstruct *slips = (slipstruct *) NULL;

static short
halfrandom(slipstruct *sp, int mv)
{
	unsigned long r;

	if (sp->lasthalf) {
		r = sp->lasthalf;
		sp->lasthalf = 0;
	}
	else {
		r = LRAND();
		sp->lasthalf = (short) (r >> 16);
	}
	return r % mv;
}

static int
erandom(slipstruct *sp, int mv)
{
	int         res;

	if (0 == sp->stage) {
		sp->r = LRAND();
		sp->stage = 7;
	}
	res = (int) (sp->r & 0xf);
	sp->r = sp->r >> 4;
	sp->stage--;
	if (res & 8)
		return res & mv;
	else
		return -(res & mv);
}

static void
prepare_screen(slipstruct *sp)
{

	int         i, n, w = sp->width / 20;

	sp->backwards = (int) (LRAND() & 1);	/* jwz: go the other way sometimes */

	if (sp->first_time || !halfrandom(sp, 10)) {
		//draw(screen, screen->r, display->black, nil, ZP);
		n = 300;
	}
	else {
		if (halfrandom(sp, 5))
			return;
		if (halfrandom(sp, 5))
			n = 100;
		else
			n = 2000;
	}

	if(LRAND() & 1) {
		draw(screen, screen->r, img, nil, ZP);
	}
	else {

		for (i = 0; i < n; i++) {
			Point 	p[4];
			int         ww = ((w / 2) + halfrandom(sp, MAX(w, 1)));

			p[0] = addpt(screen->r.min, Pt(halfrandom(sp, MAX(sp->width - ww, 1)), halfrandom(sp, MAX(sp->height - ww, 1))));
			p[1] = addpt(p[0], Pt(0, ww));
			p[2] = addpt(p[0], Pt(ww, ww));
			p[3] = addpt(p[0], Pt(ww, 0));

			fillpoly(screen, p, 4, 0, colors[halfrandom(sp, 255)], ZP);
		}
	}
	sp->first_time = 0;
}

static int
quantize(double d)
{
	int         i = (int) floor(d);
	double      f = d - i;

	if ((LRAND() & 0xff) < f * 0xff)
		i++;
	return i;
}

void
init_slip(void)
{
	slipstruct *sp;

	if (slips == NULL) {
		if ((slips = (slipstruct *) calloc(1,
			sizeof (slipstruct))) == NULL)
			return;
	}
	sp = &slips[0];

	sp->width = Dx(screen->r);
	sp->height = Dy(screen->r);

	sp->blit_width = sp->width / 25;
	sp->blit_height = sp->height / 25;
	sp->nblits_remaining = 0;
	sp->mode = 0;
	sp->first_time = 1;


}

void
draw_slip(void)
{
	int         timer;
	slipstruct *sp;

	if (slips == NULL)
		return;
	sp = &slips[0];

	timer = 1000 * 10;


	while (timer--) {
		int         xi = halfrandom(sp, MAX(sp->width - sp->blit_width, 1));
		int         yi = halfrandom(sp, MAX(sp->height - sp->blit_height, 1));
		double      x, y, dx = 0, dy = 0, t, s1, s2;

		if (0 == sp->nblits_remaining--) {
			static int  lut[] =
			{
				0, 0, 0, 1, 1, 1, 2						};


			prepare_screen(sp);
			sp->nblits_remaining = 10 *
				(2000 + halfrandom(sp, 1000) + halfrandom(sp, 1000));
			if (sp->mode == 2)
				sp->mode = halfrandom(sp, 2);
			else
				sp->mode = lut[halfrandom(sp, 7)];
		}
		x = (2 * xi + sp->blit_width) / (double) sp->width - 1;
		y = (2 * yi + sp->blit_height) / (double) sp->height - 1;

		/* (x,y) is in biunit square */
		switch (sp->mode) {
		case 0:	/* rotor */
			dx = x;
			dy = y;

			if (dy < 0) {
				dy += 0.04;
				if (dy > 0)
					dy = 0.00;
			}
			if (dy > 0) {
				dy -= 0.04;
				if (dy < 0)
					dy = 0.00;
			}
			t = dx * dx + dy * dy + 1e-10;
			s1 = 2 * dx * dx / t - 1;
			s2 = 2 * dx * dy / t;
			dx = s1 * 5;
			dy = s2 * 5;

			if (sp->backwards) {	/* jwz: go the other way sometimes */
				dx = -dx;
				dy = -dy;
			}
			break;
		case 1:	/* shuffle */
			dx = erandom(sp, 3);
			dy = erandom(sp, 3);
			break;
		case 2:	/* explode */
			dx = x * 3;
			dy = y * 3;
			break;
		}
		{
			int         qx = xi + quantize(dx), qy = yi + quantize(dy);
			int         wrap;
			Rectangle r;
			Point p, p2;

			if (qx < 0 || qy < 0 ||
				qx >= sp->width - sp->blit_width ||
				qy >= sp->height - sp->blit_height)
				continue;

			p = addpt(screen->r.min, Pt(qx, qy));
			p2 = addpt(screen->r.min, Pt(xi, yi));
			r = Rpt(p, addpt(Pt(sp->blit_width, sp->blit_height), p));
			draw(screen, r, screen, nil, p2);
			//			XCopyArea(display, window, window, gc, xi, yi,
			//				  sp->blit_width, sp->blit_height,
			//				  qx, qy);
			switch (sp->mode) {
			case 0:
				/* wrap */
				wrap = sp->width - (2 * sp->blit_width);
				if (qx > wrap ) {
					p = addpt(screen->r.min, Pt(qx - wrap, qy));
					p2 = addpt(screen->r.min, Pt(qx, qy));
					r = Rpt(p, addpt(Pt(sp->blit_width, sp->blit_height), p));
					draw(screen, r, screen, nil, p2);
					//						XCopyArea(display, window, window, gc, qx, qy,
					//						sp->blit_width, sp->blit_height,
					//							  qx - wrap, qy);
				}
				if (qx < 2 * sp->blit_width) {
					p = addpt(screen->r.min, Pt(qx + wrap, qy));
					p2 = addpt(screen->r.min, Pt(qx, qy));
					r = Rpt(p, addpt(Pt(sp->blit_width, sp->blit_height), p));
					draw(screen, r, screen, nil, p2);
					//						XCopyArea(display, window, window, gc, qx, qy,
					//						sp->blit_width, sp->blit_height,
					//							  qx + wrap, qy);
				}
				wrap = sp->height - (2 * sp->blit_height);
				if (qy > wrap) {
					p = addpt(screen->r.min, Pt(qx, qy - wrap));
					p2 = addpt(screen->r.min, Pt(qx, qy));
					r = Rpt(p, addpt(Pt(sp->blit_width, sp->blit_height), p));
					draw(screen, r, screen, nil, p2);
					//						XCopyArea(display, window, window, gc, qx, qy,
					//						sp->blit_width, sp->blit_height,
					//							  qx, qy - wrap);
				}
				if (qy < 2 * sp->blit_height) {
					p = addpt(screen->r.min, Pt(qx, qy + wrap));
					p2 = addpt(screen->r.min, Pt(qx, qy));
					r = Rpt(p, addpt(Pt(sp->blit_width, sp->blit_height), p));
					draw(screen, r, screen, nil, p2);
					//						XCopyArea(display, window, window, gc, qx, qy,
					//						sp->blit_width, sp->blit_height,
					//							  qx, qy + wrap);
				}
				break;
			case 1:
			case 2:
				break;
			}
		}
	}
	flushimage(display, 1);
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window: %r");
	}

	if(img != nil)
		freeimage(img);

	img = allocimage(display, Rect(0, 0, Dx(screen->r), Dy(screen->r)), 
		screen->chan, 0, DBlack);
	img = squeeze(display->image, img);
	if(img == nil)
		sysfatal("unable to fit screenshot to window: %r");

	init_slip();
	draw(screen, screen->r, img, nil, ZP);
}



void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));

	if(initdraw(nil, nil, "slip") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(65535)));
	}
	einit(Emouse);
	eresized(0);

	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_slip();
		frames++;
	}
}
