/* 
 * create a smooth colormap and fit it in an array
 * of 256 Images. 
 * 
 * Idea stolen from xscreensaver's swirl.c
 */

enum {
	Maxcolors = 256,
};

typedef struct SmColor SmColor;
struct SmColor {
	ushort r;
	ushort g;
	ushort b;
};

static void
interpolate(Image *colors[], ulong *pixel, int n, SmColor *c1, SmColor *c2)
{
	int         i, r, g, b;
	SmColor      c;
	unsigned short maxv;

	/* maximum value */
	maxv = Maxcolors - 1;

	for (i = 0; i < n / 2; i++) {
		/* work out the colour */
		r = c1->r + 2 * i * ((int) c2->r) / n;
		c.r = (r > (int) maxv) ? maxv : r;
		g = c1->g + 2 * i * ((int) c2->g) / n;
		c.g = (g > (int) maxv) ? maxv : g;
		b = c1->b + 2 * i * ((int) c2->b) / n;
		c.b = (b > (int) maxv) ? maxv : b;

		/* set it up */
		colors[(*pixel)++] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			c.r << 24 | c.g << 16 | c.b << 8 | 0xff);
	}
	for (i = n / 2; i >= 0; i--) {
		r = c2->r + 2 * i * ((int) c1->r) / n;
		c.r = (r > (int) maxv) ? maxv : r;
		g = c2->g + 2 * i * ((int) c1->g) / n;
		c.g = (g > (int) maxv) ? maxv : g;
		b = c2->b + 2 * i * ((int) c1->b) / n;
		c.b = (b > (int) maxv) ? maxv : b;

		/* set it up */
		colors[(*pixel)++] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			c.r << 24 | c.g << 16 | c.b << 8 | 0xff);
	}
}


void
smooth(Image *colors[])
{
	int i; 
	unsigned short r1, g1, b1, r2, g2, b2, r3, g3, b3;
	int         L1, L2, L3, L;
	SmColor c[3];
	ulong pixel = 0;

	for(i = 0; i < Maxcolors; i++)
		if(colors[i])
			freeimage(colors[i]);

	for(i = 0; i < 3; i++) {
		c[i].r = nrand(Maxcolors);
		c[i].g = nrand(Maxcolors);
		c[i].b = nrand(Maxcolors);
	}

	r1 = c[0].r;
	g1 = c[0].g;
	b1 = c[0].b;
	r2 = c[1].r;
	g2 = c[1].g;
	b2 = c[1].b;
	r3 = c[2].r;
	g3 = c[2].g;
	b3 = c[2].b;

	/* work out the lengths of each side of the triangle */
	L1 = (int) sqrt((((double) r1 - r2) * ((double) r1 - r2) +
			 ((double) g1 - g2) * ((double) g1 - g2) +
			 ((double) b1 - b2) * ((double) b1 - b2)));

	L2 = (int) sqrt((((double) r3 - r2) * ((double) r3 - r2) +
			 ((double) g3 - g2) * ((double) g3 - g2) +
			 ((double) b3 - b2) * ((double) b3 - b2)));

	L3 = (int) sqrt((((double) r1 - r3) * ((double) r1 - r3) +
			 ((double) g1 - g3) * ((double) g1 - g3) +
			 ((double) b1 - b3) * ((double) b1 - b3)));

	L = L1 + L2 + L3;

	interpolate(colors, &pixel, (int) (250.0 * ((double) L1 / (double) L)) + 1, c, c + 1);
	interpolate(colors, &pixel, (int) (250.0 * ((double) L2 / (double) L)) + 1, c + 1, c + 2);
	interpolate(colors, &pixel, (int) (250.0 * ((double) L3 / (double) L)) + 1, c + 2, c);

	while ((int) pixel < Maxcolors) {
		/* repeat the last colour */
		colors[pixel++] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			c[0].r << 24 | c[0].g << 16 | c[0].b << 8 | 0xff);
	}

}

static void
intcmap(uchar *buf, ulong *pixel, int n, SmColor *c1, SmColor *c2)
{
	int         i, r, g, b;
	SmColor      c;
	unsigned short maxv;

	/* maximum value */
	maxv = Maxcolors - 1;

	for (i = 0; i < n / 2; i++) {
		/* work out the colour */
		r = c1->r + 2 * i * ((int) c2->r) / n;
		c.r = (r > (int) maxv) ? maxv : r;
		g = c1->g + 2 * i * ((int) c2->g) / n;
		c.g = (g > (int) maxv) ? maxv : g;
		b = c1->b + 2 * i * ((int) c2->b) / n;
		c.b = (b > (int) maxv) ? maxv : b;

		/* set it up */
		buf[(*pixel)++] = rgb2cmap(c.r,  c.g, c.b);
	}
	for (i = n / 2; i >= 0; i--) {
		r = c2->r + 2 * i * ((int) c1->r) / n;
		c.r = (r > (int) maxv) ? maxv : r;
		g = c2->g + 2 * i * ((int) c1->g) / n;
		c.g = (g > (int) maxv) ? maxv : g;
		b = c2->b + 2 * i * ((int) c1->b) / n;
		c.b = (b > (int) maxv) ? maxv : b;

		/* set it up */
		buf[(*pixel)++] = rgb2cmap(c.r,  c.g, c.b);
	}
}


void
smoothcmap(uchar *buf)
{
	int i; 
	unsigned short r1, g1, b1, r2, g2, b2, r3, g3, b3;
	int         L1, L2, L3, L;
	SmColor c[3];
	ulong pixel = 0;

	for(i = 0; i < 3; i++) {
		c[i].r = nrand(Maxcolors);
		c[i].g = nrand(Maxcolors);
		c[i].b = nrand(Maxcolors);
	}

	r1 = c[0].r;
	g1 = c[0].g;
	b1 = c[0].b;
	r2 = c[1].r;
	g2 = c[1].g;
	b2 = c[1].b;
	r3 = c[2].r;
	g3 = c[2].g;
	b3 = c[2].b;

	/* work out the lengths of each side of the triangle */
	L1 = (int) sqrt((((double) r1 - r2) * ((double) r1 - r2) +
			 ((double) g1 - g2) * ((double) g1 - g2) +
			 ((double) b1 - b2) * ((double) b1 - b2)));

	L2 = (int) sqrt((((double) r3 - r2) * ((double) r3 - r2) +
			 ((double) g3 - g2) * ((double) g3 - g2) +
			 ((double) b3 - b2) * ((double) b3 - b2)));

	L3 = (int) sqrt((((double) r1 - r3) * ((double) r1 - r3) +
			 ((double) g1 - g3) * ((double) g1 - g3) +
			 ((double) b1 - b3) * ((double) b1 - b3)));

	L = L1 + L2 + L3;

	intcmap(buf, &pixel, (int) (250.0 * ((double) L1 / (double) L)) + 1, c, c + 1);
	intcmap(buf, &pixel, (int) (250.0 * ((double) L2 / (double) L)) + 1, c + 1, c + 2);
	intcmap(buf, &pixel, (int) (250.0 * ((double) L3 / (double) L)) + 1, c + 2, c);

	while ((int) pixel < Maxcolors) {
		/* repeat the last colour */
		buf[pixel++] = rgb2cmap(c[0].r,c[0].g,c[0].b);
	}

}

void
randomcmap(Image *colors[])
{
	int i; 

	for(i = 0; i < Maxcolors; i++) {
		if(colors[i])
			freeimage(colors[i]);
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, 
			cmap2rgba(nrand(Maxcolors)));
	}
}
