/* The Spiral Generator, Copyright (c) 2000 
 * by Rohit Singh <rohit_singh@hotmail.com>
 * 
 * Contains code from / To be used with:
 * xscreensaver, Copyright (c) 1992, 1995, 1996, 1997
 * Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notices appear in all copies and that both that
 * copyright notices and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 *
 * Modified (Dec 2001) by Matthew Strait <straitm@mathcs.carleton.edu>
 * Added -subdelay and -alwaysfinish
 * Prevented redrawing over existing lines
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 06/02
  */


/* plan9-related stuff */
#include "plan9.h"


static unsigned int default_fg_pixel;

static void
init_tsg (void)
{

}


static void
go (int radius1, int radius2,
int d)
{
	int width, height;
	int xmid, ymid;
	int x1, y1, x2, y2;
	float firstx = 0, firsty = 0;
	float tmpx, tmpy;
	int theta;
	int delta;

	width  = Dx(screen->r);
	height = Dy(screen->r);
	delta = 1;
	xmid = width / 2;
	ymid = height / 2;

	x1 = xmid + radius1 - radius2 + d;
	y1 = ymid;

	for (theta = 1; /* theta < ( 360 * 100 ) */; theta++)
	{
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}

		tmpx = xmid + ((       radius1	         
			- radius2        )
			* cos((      theta 		
			* M_PI           )
			/ 180           ))
			+ (              d 		
			* cos((((  radius1 		
			* theta          )
			- delta          )
			/ radius2        )
			*             M_PI		
			/ 180            )
			);
		tmpy = ymid + (				
			( radius1 - radius2	
			) * sin			
			(			
			( theta * M_PI     	
			) / 180		
			)
			) +                  
			( d * sin           
			(                  
			(                 
			(          	
			( radius1 * theta
			) - delta
			) / radius2
			) * M_PI / 180
			)
			);

		/*makes integers from the calculated values to do the drawing*/
		x2 = tmpx;
		y2 = tmpy;

		/*stores the first values for later reference*/
		if(theta == 1)
		{
			firstx = tmpx;
			firsty = tmpy;
		}

		//        XDrawLine (dpy, window, draw_gc, x1, y1, x2, y2);
		line(screen, Pt(screen->r.min.x + x1, screen->r.min.y + y1), 
			Pt(screen->r.min.x + x2, screen->r.min.y + y2), 
			Endsquare, Endsquare, 0, color, ZP);
		flushimage(display, 1);
		frames++;

		x1 = x2;
		y1 = y2;

		if(tmpx == firstx && tmpy == firsty && theta != 1)
			break;

		if(theta > ( 360 * 1000 ) )
			break;

	}
}



static void
getset (void)
{
	int width, height;
	int radius, radius1, radius2;
	double divisor;
	int distance;

	width = Dx(screen->r);
	height = Dy(screen->r);

	radius = MIN(width, height) / 2;

	draw(screen, screen->r, display->black, nil, ZP);

	divisor = ((frand()*3.0 + 1) * (((lrand() & 1) * 2) - 1));

	radius1 = radius;
	radius2 = radius / divisor + 5;
	distance = 100 + (lrand() % 200);

	color = colors[nrand(256)];
	go (radius1, -radius2, distance);
	color = colors[nrand(256)];
	go (radius1, radius2, distance);
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);


}

void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "spirograph") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++)
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(65535)));

	einit(Emouse);

	eresized(0);
	for(;;) {
		getset();
	}
}
