/* squiral, by "Jeff Epler" <jepler@inetnebr.com>, 18-mar-1999.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 07/02
 * fixed for proper colors: 09/03
 */

/* plan9-related stuff */
#include "plan9.h"

#define R(x)  (abs(random())%x)
#define PROB(x) (frand() < (x))

#define NCOLORSMAX 256
static int width, height, count, cycle;
static double frac, disorder, handedness;
static int ncolors=0;
#define STATES 8

/* 0- 3 left-winding  */
/* 4- 7 right-winding */

#define CLEAR1(x,y) (!fill[((y)%height)*width+(x)%width])

static int cov;
static int dirh[4];
static int dirv[4];

static int *fill;

#define CLEARDXY(x,y,dx,dy) CLEAR1(x+dx, y+dy) && CLEAR1(x+dx+dx, y+dy+dy)
#define MOVEDXY(x,y,dx,dy)  MOVE1 (x+dx, y+dy), MOVE1 (x+dx+dx, y+dy+dy)

#define CLEAR(d) CLEARDXY(w->h,w->v, dirh[d],dirv[d])
#define MOVE(d) (color = colors[w->c], \
                  MOVEDXY(w->h,w->v, dirh[d],dirv[d]), \
		  w->h=w->h+dirh[d]*2, \
		  w->v=w->v+dirv[d]*2, dir=d)

#define RANDOM  (w->h = R(width), w->v = R(height), w->c = R(ncolors), \
		  type=R(2), dir=R(4), (cycle && (w->cc=R(3)+ncolors)))

struct worm {
	int h;
	int v;
	int s;
	int c;
	int cc;
} *worms;

#define SUCC(x) ((x+1)%4)
#define PRED(x) ((x+3)%4)
#define CCW     PRED(dir)
#define CW      SUCC(dir)
#define REV     ((dir+2)%4)
#define STR     (dir)
#define TRY(x)  if (CLEAR(x)) { MOVE(x); break; }

static void
MOVE1(int x, int y)
{
	Rectangle r;

	fill[((y)%height)*width+(x)%width]=1;

	r = Rpt(addpt(screen->r.min, Pt(x%width, y%height)), addpt(screen->r.min, Pt(x%width+1, y%height+1)));

	draw(screen, r, color, nil, ZP);
}

static void
do_worm(struct worm *w)
{
	int type = w->s / 4;
	int dir = w->s % 4;

	w->c = (w->c+w->cc) % ncolors;

	if (PROB(disorder)) type=PROB(handedness);
	switch(type) {
	case 0: /* CCW */
		TRY(CCW)
			TRY(STR)
			TRY(CW)
			RANDOM;
		break;
	case 1: /* CW */
		TRY(CW)
			TRY(STR)
			TRY(CCW)
			RANDOM;
		break;
	}
	w->s = type*4+dir;
	w->h = w->h % width;
	w->v = w->v % height;
}

static void
init_squiral(void)
{
	int i;

	width  = Dx(screen->r);
	height = Dy(screen->r);

	ncolors = NCOLORSMAX;

	cycle= False;
	disorder=0.005;
	handedness=0.5;
	frac = frand();
	count=width/32;

	if(worms) free(worms);
	if(fill)  free(fill);

	worms=calloc(count, sizeof(struct worm));
	fill=calloc(width*height, sizeof(int));

	dirh[0]=0;
	dirh[1]=1;
	dirh[2]=0;
	dirh[3]=width-1;
	dirv[0]=height-1;
	dirv[1]=0;
	dirv[2]=1;
	dirv[3]=0;
	for(i=0;i<count;i++) {
		worms[i].h=R(width);
		worms[i].v=R(height);
		worms[i].s=R(4)+4*PROB(handedness);
		worms[i].c=R(ncolors);
		if(cycle) {
			worms[i].cc=R(3)+ncolors;
		}
		else worms[i].cc=0;
	}
}

int init = 0;

void
screenhack(void)
{
	int inclear;
	int i;
	int counter = 0;
	init_squiral();
	cov=0;
	inclear=height;
	while(1) {
		if(inclear<height) {
			//XDrawLine(dpy, window, erase_gc, 0, inclear, width-1, inclear);
			line(screen, addpt(screen->r.min, Pt(0, inclear)), 
				addpt(screen->r.min, Pt(width-1, inclear)), 
				Endsquare, Endsquare, 0, display->black, ZP);

			memset(&fill[inclear*width], 0, sizeof(int)*width);
			//XDrawLine(dpy, window, erase_gc, 0, height-inclear-1, width-1,
			//    height-inclear-1);
			line(screen, addpt(screen->r.min, Pt(0, height-inclear-1)), 
				addpt(screen->r.min, Pt(width-1, height-inclear-1)), 
				Endsquare, Endsquare, 0, display->black, ZP);

			memset(&fill[(height-inclear-1)*width], 0, sizeof(int)*width);
			inclear++;
			//XDrawLine(dpy, window, erase_gc, 0, inclear, width-1, inclear);
			line(screen, addpt(screen->r.min, Pt(0, inclear)), 
				addpt(screen->r.min, Pt(width-1, inclear)), 
				Endsquare, Endsquare, 0, display->black, ZP);

			memset(&fill[inclear*width], 0, sizeof(int)*width);
			//XDrawLine(dpy, window, erase_gc, 0, height-inclear-1, width-1,
			//    height-inclear-1);
			line(screen, addpt(screen->r.min, Pt(0, height-inclear-1)), 
				addpt(screen->r.min, Pt(width-1, height-inclear-1)), 
				Endsquare, Endsquare, 0, display->black, ZP);

			memset(&fill[(height-inclear-1)*width], 0, sizeof(int)*width);
			inclear++;
			if(inclear>height/2) inclear=height;
		}
		else if(cov>(frac*width*height)) {
			inclear=0;
			cov=0;
		}

		for(i=0;i<count;i++) do_worm(&worms[i]);

		flushimage(display, 1);
		frames++;
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if(counter++ > 5000)
			init = 1, counter = 0;

		if(init) {
			init = 0;
			draw(screen, screen->r, display->black, nil, ZP);
			init_squiral();
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	init = 1;
	draw(screen, screen->r, display->black, nil, ZP);
}

void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "squiral") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);
	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(256)));
	}
	color = colors[nrand(256)];

	eresized(0);
	screenhack();
}
