/*
 *  Substrate (dragorn@kismetwireless.net)
 *  Directly ported code from complexification.net Substrate art
 *  http://complexification.net/gallery/machines/substrate/applet_s/substrate_s.pde
 *
 *  Substrate code:
 *  j.tarbell   June, 2004
 *  Albuquerque, New Mexico
 *  complexification.net
 *
 *  CHANGES
 *
 *  1.0  dragorn  Oct 10 2004    First port done
 *
 * Directly based the hacks of: 
 * 
 * xscreensaver, Copyright (c) 1997, 1998, 2002 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchovski@gmail.com, 01/05
 */

#include "plan9.h"
#include <memdraw.h>
#include <memlayer.h>

ulong rgbc[] = {
	0x201F21FF, 0x262C2EFF, 0x352626FF, 0x372B27FF,
		0x302C2EFF, 0x392B2DFF, 0x323229FF, 0x3F3229FF,
		0x38322EFF, 0x2E333DFF, 0x333A3DFF, 0x473329FF,
		0x40392CFF, 0x40392EFF, 0x47402CFF, 0x47402EFF,
		0x4E402CFF, 0x4F402EFF, 0x4E4738FF, 0x584037FF,
		0x65472DFF, 0x6D5D3DFF, 0x745530FF, 0x755532FF,
		0x745D32FF, 0x746433FF, 0x7C6C36FF, 0x523152FF,
		0x444842FF, 0x4C5647FF, 0x655D45FF, 0x6D5D44FF,
		0x6C5D4EFF, 0x746C43FF, 0x7C6C42FF, 0x7C6C4BFF,
		0x6B734BFF, 0x73734BFF, 0x7B7B4AFF, 0x6B6C55FF,
		0x696D5EFF, 0x7B6C5DFF, 0x6B7353FF, 0x6A745DFF,
		0x727B52FF, 0x7B7B52FF, 0x57746EFF, 0x687466FF,
		0x9C542BFF, 0x9D5432FF, 0x9D5B35FF, 0x936B36FF,
		0xAA7330FF, 0xC45A27FF, 0xD95223FF, 0xD85A20FF,
		0xDB5A23FF, 0xE57037FF, 0x836C4BFF, 0x8C6B4BFF,
		0x82735CFF, 0x937352FF, 0x817B63FF, 0x817B6DFF,
		0x927B63FF, 0xD9893BFF, 0xE49832FF, 0xDFA133FF,
		0xE5A037FF, 0xF0AB3BFF, 0x8A8A59FF, 0xB29A58FF,
		0x89826BFF, 0x9A8262FF, 0x888B7CFF, 0x909A7AFF,
		0xA28262FF, 0xA18A69FF, 0xA99968FF, 0x99A160FF,
		0x99A168FF, 0xCA8148FF, 0xEB8D43FF, 0xC29160FF,
		0xC29168FF, 0xD1A977FF, 0xC9B97FFF, 0xF0E27BFF,
		0x9F928BFF, 0xC0B999FF, 0xE6B88FFF, 0xC8C187FF,
		0xE0C886FF, 0xF2CC85FF, 0xF5DA83FF, 0xECDE9DFF,
		0xF5D294FF, 0xF5DA94FF, 0xF4E784FF, 0xF4E18AFF,
		0xF4E193FF, 0xE7D8A7FF, 0xF1D4A5FF, 0xF1DCA5FF,
		0xF4DBADFF, 0xF1DCAEFF, 0xF4DBB5FF, 0xF5DBBDFF,
		0xF4E2ADFF, 0xF5E9ADFF, 0xF4E3BEFF, 0xF5EABEFF,
		0xF7F0B6FF, 0xD9D1C1FF, 0xE0D0C0FF, 0xE7D8C0FF,
		0xF1DDC6FF, 0xE8E1C0FF, 0xF3EDC7FF, 0xF6ECCEFF,
		0xF8F2C7FF, 0xEFEFD0FF, 0
};

typedef struct {
	float x, y;
	float t;

	ulong sandcolor;
	float sandp, sandg;
} crack;

struct field {
	uint height;
	uint width;

	uint initial_cracks;

	uint num;
	uint max_num;

	int grains; /* number of grains in the sand painting */

	crack *cracks; /* grid of cracks */
	int *cgrid; /* grid of actual crack placement */

	/* Raw map of pixels we need to keep for alpha blending */
	ulong *img;
	Image *cimg;

	/* color parms */
	int numcolors;
	uint cycles;

	uint wireframe;
};
struct field *f;

static void 
*xrealloc(void *p, uint size)
{
	void *ret;
	if ((ret = realloc(p, size)) == nil)
		sysfatal("out of memory");
	return ret;
}

struct field 
*init_field(void)
{
	struct field *f = xrealloc(nil, sizeof(struct field));
	f->height = 0;
	f->width = 0;
	f->initial_cracks = 0;
	f->num = 0;
	f->max_num = 0;
	f->cracks = nil;
	f->cgrid = nil;
	f->img = nil;
	f->cimg = nil;
	f->numcolors = 122;
	f->cycles = 0;
	f->wireframe = 0;
	f->grains = 0;
	return f;
}

void 
start_crack(struct field *f, crack *cr)
{
	int px = 0;
	int py = 0;
	int found = 0;
	int timeout = 0;
	float a;

	/* shift until crack is found */
	while ((!found) || (timeout++ > 10000)) {
		px = (int) (nrand(f->width));
		py = (int) (nrand(f->height));

		if ( f->cgrid[py * f->width + px] < 10000)
			found = 1;
	}

	if (found) {
		/* start a crack */
		a =  f->cgrid[py * f->width + px];

		if ((random() % 100) < 50) {
			/* conversion of the java int(random(-2, 2.1)) */
			a -= 90 + (frand()*4.1 - 2.0);
		} else {
			a += 90 + (frand()*4.1 - 2.0);
		}

		/* Condensed from Crack::startCrack */
		cr->x = px + ((float) 0.61 * cos(a * M_PI / 180));
		cr->y = py + ((float) 0.61 * sin(a * M_PI / 180));
		cr->t = a;
	} else {
		/* timeout */
	}
}

void 
make_crack(struct field *f)
{
	crack *cr;

	if (f->num < f->max_num) {
		/* make a new crack */
		f->cracks = (crack *) xrealloc(f->cracks, sizeof(crack) * (f->num + 1));

		cr = &(f->cracks[f->num]);
		/* assign colors */
		cr->sandp = 0;
		cr->sandg = (frand()*0.2 - 0.01);
		cr->sandcolor = rgbc[nrand(f->numcolors)];

		/* start it */
		start_crack(f, cr);

		f->num++;
	}
}

/* alpha blended point drawing */
void 
trans_point(int x1, int y1, ulong myc, float a, struct field *f)
{
	if ((x1 >= 0) && (x1 < f->width) && (y1 >= 0) && (y1 < f->height)) {
		if (a >= 1.0) {
			f->img[y1*f->width + x1] = myc;
		} else {
			int or, og, ob;
			int r, g, b;
			int nr, ng, nb;
			ulong c;

			c = f->img[y1*f->width + x1];

			or = (c >> 24) & 0xff;
			og = (c >> 16) & 0xff;
			ob = (c >> 8) & 0xff;

			r = (myc >> 24) & 0xff;
			g = (myc >> 16) & 0xff;
			b = (myc >> 8) & 0xff;

			nr = or + (r - or) * a;
			ng = og + (g - og) * a;
			nb = ob + (b - ob) * a;

			c = (nr << 24) | (ng << 16) | (nb << 8) | 0xff;
			f->img[y1*f->width + x1] = c;
		}
	}
}

void 
region_color(struct field *f, crack *cr)
{

	float rx = cr->x;
	float ry = cr->y;
	int openspace = 1;
	int cx, cy;
	float maxg;
	int grains, i;
	float w;
	float drawx, drawy;

	while (openspace) {
		/* move perpendicular to crack */
		rx += (0.81 * sin(cr->t * M_PI/180));
		ry -= (0.81 * cos(cr->t * M_PI/180));

		cx = (int) rx;
		cy = (int) ry;

		if ((cx >= 0) && (cx < f->width) && (cy >= 0) && (cy < f->height)) {
			/* safe to check */
			if (f->cgrid[cy * f->width + cx] > 10000) {
				/* space is open */
			} else {
				openspace = 0;
			}
		} else {
			openspace = 0;
		}
	}

	/* SandPainter stuff here */

	/* Modulate gain */
	cr->sandg += (frand()*0.1 - 0.050);
	maxg = 1.0;

	if (cr->sandg < 0)
		cr->sandg = 0;

	if (cr->sandg > maxg)
		cr->sandg = maxg;

	grains = f->grains;

	/* Lay down grains of sand */
	w = cr->sandg / (grains - 1);

	for (i = 0; i < grains; i++) {
		drawx = (cr->x + (rx - cr->x) * sin(cr->sandp + sin((float) i * w)));
		drawy = (cr->y + (ry - cr->y) * sin(cr->sandp + sin((float) i * w)));

		/* Draw sand bit */
		trans_point((int)drawx+0.5, (int)drawy+0.5, cr->sandcolor, (0.1 - i / (grains * 10.0)), f);
	}
}

void 
build_substrate(struct field *f)
{
	int tx, ty;

	f->cycles = 0;

	if (f->cgrid) {
		free(f->cgrid);
		f->cgrid = NULL;
	}

	if (f->cracks) {
		free(f->cracks);
		f->cracks = NULL;
	}

	f->num = 0;

	/* erase the crack grid */
	f->cgrid = (int *) xrealloc(f->cgrid, sizeof(int) * f->height * f->width);
	memset(f->cgrid, 10001, f->height * f->width * sizeof(int));

	/* make random crack seeds */
	for (tx = 0; tx < 16; tx++) {
		ty = (int) (random() % (f->width * f->height - 1));
		f->cgrid[ty] = (int) random() % 360;
	}

	/* make the initial cracks */
	for (tx = 0; tx < f->initial_cracks; tx++)
		make_crack(f);
}


void 
movedrawcrack(struct field *f, int cracknum)
{
	int cx, cy;
	crack *cr = &(f->cracks[cracknum]);

	/* continue cracking */
	cr->x += ((float) 0.42 * cos(cr->t * M_PI/180));
	cr->y += ((float) 0.42 * sin(cr->t * M_PI/180));

	/* bounds check */
	/* modification of random(-0.33,0.33) */
	cx = (int) (cr->x + (frand()*0.66 - 0.33));
	cy = (int) (cr->y + (frand()*0.66 - 0.33));

	if ((cx >= 0) && (cx < f->width) && (cy >= 0) && (cy < f->height)) {
		/* draw sand painter if we're not wireframe */
		if (!f->wireframe)
			region_color(f, cr);
		f->img[cy*f->width + cx] = 0xff;

		/* safe to check */
		if ((f->cgrid[cy * f->width + cx] > 10000) ||
			(abs(f->cgrid[cy * f->width + cx] - cr->t) < 5)) {
			/* continue cracking */
			f->cgrid[cy * f->width + cx] = (int) cr->t;
		} else if (abs(f->cgrid[cy * f->width + cx] - cr->t) > 2) {
			/* crack encountered (not self), stop cracking */
			start_crack(f, cr); /* restart ourselves */
			make_crack(f); /* generate a new crack */
		}
	} else {
		/* out of bounds, top cracking */
		start_crack(f, cr); /* restart ourselves */
		make_crack(f); /* generate a new crack */
	}
}

char *progclass = "Substrate";
void 
screenhack(void)
{

	uint max_cycles = 2000;
	int tempx;

	while (1) {
		for (tempx = 0; tempx < f->num; tempx++) {
			movedrawcrack(f, tempx);
		}
		if(f->cycles % 5 == 0) {
			loadimage(f->cimg, f->cimg->r, (uchar *)f->img, f->height*f->width*sizeof(ulong));
			draw(screen, screen->r, f->cimg, nil, ZP);
			flushimage(display, 1);
			frames++;
		}

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(nil);
		}

		if (f->cycles++ > max_cycles) {
			build_substrate(f);
			memset(f->img, 0xffffffff, f->width*f->height*sizeof(ulong));
			draw(f->cimg, f->cimg->r, display->white, nil, ZP);
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	f->width = Dx(screen->r);
	f->height = Dy(screen->r);

	build_substrate(f);
	if(f->img != nil)
		free(f->img);
	f->img = malloc(f->width*f->height*sizeof(ulong));
	if(f->img == nil)
		sysfatal("img malloc");
	memset(f->img, 0xffffffff, f->width*f->height*sizeof(ulong));

	if(f->cimg != nil)
		freeimage(f->cimg);
	f->cimg = allocimage(display, Rect(0, 0, f->width, f->height), RGBA32, 1, DWhite);
	if(f->cimg == nil)
		sysfatal("cimg allocimage");
	draw(f->cimg, f->cimg->r, display->white, nil, ZP);
}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "substrate") < 0)
		sysfatal("initdraw failed: %r");

	memimageinit();

	einit(Emouse);

	f = init_field();
	f->initial_cracks = 3;
	f->max_num = 100;
	f->wireframe = 0; //nrand(2);
	f->grains = 64;

	eresized(0);

	screenhack();
}
