/* vines --- vine fractals */


/*-
 * Copyright (c) 1997 by Tracy Camp campt@hurrah.com
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * If you make a modification I would of course appreciate a copy.
 *
 * Revision History:
 * 01-Nov-2000: Allocation checks
 * 11-Jul-1997: David Hansen <dhansen@metapath.com>
 *              Changed names to vines and modified draw loop
 *              to honor batchcount so vines can be grown or plotted.
 * 10-May-1997: Compatible with xscreensaver
 * 21-Mar-1997: David Hansen <dhansen@metapath.com>
 *              Updated mode to draw complete patterns on every
 *              iteration instead of growing the vine.  Also made
 *              adjustments to randomization and changed variable
 *              names to make logic easier to follow.
 */

/*-
 * This was modifed from a 'screen saver' that a friend and I
 * wrote on our TI-8x calculators in high school physics one day
 * Basically another geometric pattern generator, this ones claim
 * to fame is a pseudo-fractal looking vine like pattern that creates
 * nifty whorls and loops.
 */

/*
  * ported to Plan 9 by andrey@lanl.gov, 06/02
  */

/* plan9-related stuff */
#include "plan9.h"

typedef struct {
	int         a;
	int         x1;
	int         y1;
	int         x2;
	int         y2;
	int         i;
	int         length;
	int         iterations;
	int         constant;
	int         ang;
	int         centerx;
	int         centery;
} vinestruct;

static vinestruct *vines = (vinestruct *) NULL;


void
init_vines(void)
{
	vinestruct *fp;

	if (vines == NULL) {
		if ((vines = (vinestruct *) calloc(1,
			sizeof (vinestruct))) == NULL) {
			return;
		}
	}
	fp = &vines[0];

	fp->i = 0;
	fp->length = 0;
	fp->iterations = 30 + NRAND(100);

	draw(screen, screen->r, display->black, nil, ZP);
}				/* init_vines */

void
draw_vines(void)
{
	int         count;
	vinestruct *fp;
	Point p1, p2;

	if (vines == NULL)
		return;
	fp = &vines[0];

	if (fp->i >= fp->length) {
		if (--(fp->iterations) == 0) {
			draw(screen, screen->r, display->black, nil, ZP);
			init_vines();
		}
		fp->centerx = NRAND(screen->r.min.x + Dx(screen->r));
		fp->centery = NRAND(screen->r.min.y + Dy(screen->r));

		fp->ang = 60 + NRAND(720);
		fp->length = 100 + NRAND(3000);
		fp->constant = fp->length * (10 + NRAND(10));

		fp->i = 0;
		fp->a = 0;
		fp->x1 = 0;
		fp->y1 = 0;
		fp->x2 = 1;
		fp->y2 = 0;

		color = colors[nrand(255)];
	}
	count = fp->i + 2000;
	if ((count <= fp->i) || (count > fp->length))
		count = fp->length;

	while (fp->i < count) {
		p1.x = screen->r.min.x + fp->centerx + (fp->x1 / fp->constant);
		p1.y = screen->r.min.y + fp->centery - (fp->y1 / fp->constant);
		p2.x = screen->r.min.x + fp->centerx + (fp->x2 / fp->constant);
		p2.y = screen->r.min.y + fp->centery - (fp->y2 / fp->constant);

		line(screen, p1, p2, Endsquare, Endsquare, 0, color, ZP);


		fp->a += (fp->ang * fp->i);

		fp->x1 = fp->x2;
		fp->y1 = fp->y2;

		fp->x2 += (int) (fp->i * (cos((double) fp->a) * 360.0) / (2.0 * M_PI));
		fp->y2 += (int) (fp->i * (sin((double) fp->a) * 360.0) / (2.0 * M_PI));
		fp->i++;
	}
	flushimage(display, 1);
}				/* draw_vines */

void
release_vines(void)
{
	if (vines != NULL) {
		free((void *) vines);
		vines = (vinestruct *) NULL;
	}
}				/* release_vines */

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	if(vines != nil)
		release_vines();
	init_vines();
}


void 
threadmain(int argc, char **argv)
{
	Mouse m;
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "vines") < 0)
		sysfatal("initdraw failed: %r");


	for(i = 0; i < 256; i++) {
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(256)));
	}
	einit(Emouse);

	eresized(0);
	for(;;) {
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		draw_vines();
		frames++;
	}
}
