/* wander, by Rick Campbell <rick@campbellcentral.org>, 19 December 1998.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* 
  * ported to Plan 9 by andrey@lanl.gov, 06/02
  */

/* plan9-related stuff */
#include "plan9.h"

#define MAXIMUM_COLOR_COUNT (256)

static unsigned int advance     = 0;
static Bool         circles     = 0;
static int          color_count = 256;
static int          color_index = 0;
static unsigned int density     = 0;
static int          depth       = 0;
static int          height      = 0;
static unsigned int length 	= 0;
static unsigned int reset 	= 0;
static unsigned int size 	= 0;
static int          width       = 0;

static void
init_wander (void)
{

	width = Dx(screen->r);
	height = Dy(screen->r);
	depth = screen->depth;

	color_index = rand() % color_count;

	advance = 1;
	density = 2;
	if (density < 1) density = 1;
	reset = 2500000;
	if (reset < 100) reset = 100;
	circles = rand() & 1;
	size = 1;
	if (size < 1) size = 1;
	width = width / size;
	height = height / size;
	length = 25000;
	if (length < 1) length = 1;
}


static void
wander (void)
{
	int x = lrand () % width;
	int y = lrand () % height;
	int last_x = x;
	int last_y = y;
	int width_1 = width - 1;
	int height_1 = height - 1;
	int length_limit = length;
	int reset_limit = reset;
	int color_index = rand () % color_count;
	Image * color = colors [rand () % color_count];
	Mouse m;

	draw(screen, screen->r, display->black, nil, ZP);


	while (1)
	{
		if (rand () % density)
		{
			x = last_x;
			y = last_y;
		}
		else
		{
			last_x = x;
			last_y = y;
			x = (x + width_1  + (rand () % 3)) % width;
			y = (y + height_1 + (rand () % 3)) % height;
		}
		if ((rand () % length_limit) == 0)
		{
			if (advance == 0)
			{
				color_index = rand () % color_count;
			}
			else
			{
				color_index = (color_index + advance) % color_count;
			}
			color = colors [color_index];
		}
		if (rand() > reset_limit)
		{
			draw(screen, screen->r, display->black, nil, ZP);
			color = colors [rand () % color_count];
			x = rand () % width;
			y = rand () % height;
			last_x = x;
			last_y = y;

		}

		draw(screen, rectaddpt(Rect(x, y, x+1, y+1), screen->r.min), color, nil, ZP);
		flushimage(display, 1);
		frames++;

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4) {
				if(emenuhit(3, &m, &menu) == 0)
					threadexitsall(0);
			}
		}
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	draw(screen, screen->r, display->black, nil, ZP);

	init_wander();


}


void 
threadmain(int argc, char **argv)
{
	int i;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "wander") < 0)
		sysfatal("initdraw failed: %r");

	for(i = 0; i < 256; i++)
		colors[i] = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, cmap2rgba(nrand(65536)));

	einit(Emouse);
	eresized(0);
	wander();
}
