/*
 *  Copyright (C) 2000 James Macnicol
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchov@cpsc.ucalgary.ca, 10/03
 */

/* plan9-related stuff */
#include "plan9.h"

#define MINX 0.0
#define MINY 0.0
/* This should be *way* slower than the spotlight hack was */
#define X_PERIOD 45000.0
#define Y_PERIOD 36000.0

static int sizex, sizey;

static int delay;
static int pixwidth, pixheight, pixspacex, pixspacey, lensoffsetx, lensoffsety;
static Bool lenses;

Image *lscreen; 	/* local screen for double buffering */


static int tlx, tly, s;
long x = 0;
static vlong currentTimeInMs(void)
{
	return (nsec()/1000000);
}

static void init_hack(void)
{
	int nblocksx, nblocksy;

	sizex = Dx(display->image->r);
	sizey = Dy(display->image->r);

	delay = 10;
	pixheight = pixwidth = 10;
	pixspacex = 2;
	pixspacey = 2;
	lenses = nrand(2);
	lensoffsetx = 5;
	lensoffsety = lensoffsetx;

	nblocksx = (int)ceil((double)sizex / (double)(pixwidth + pixspacex));
	nblocksy = (int)ceil((double)sizey / (double)(pixheight + pixspacey));
	if (lenses)
		s = MAX((nblocksx - 1) * lensoffsetx + pixwidth, 
			(nblocksy - 1) * lensoffsety + pixheight) * 2;
	else
		s = MAX(nblocksx, nblocksy) * 2;
}

static void onestep(void)
{
	unsigned x, y, i, j;

	vlong now;
	Point p2;
	Rectangle r;

#define nrnd(x) (random() % (x))

	now = currentTimeInMs();

	/* find new x,y */
	tlx = ((1. + sin(((float)now) / X_PERIOD * 2. * M_PI))/2.0)
		* (sizex - s/2)  + MINX;
	tly = ((1. + sin(((float)now) / Y_PERIOD * 2. * M_PI))/2.0)
		* (sizey - s/2) + MINY;

	if (lenses) {
		for (x = i = 0; x < sizex; x += (pixwidth + pixspacex), ++i)
			for (y = j = 0; y < sizey; y += (pixheight + pixspacey), ++j) {
				//XCopyArea(dpy, pm /* src */, window /* dest */, window_gc,
				//    tlx + i * lensoffsetx /* src_x */, 
				//    tly + j * lensoffsety /* src_y */,
				//    pixwidth, pixheight,
				//    x /* dest_x */, y /* dest_y */);

				r = Rect(x, y, x+pixwidth, y+pixheight);

				p2 = Pt(tlx + i * lensoffsetx, tly + j * lensoffsety);
				drawop(lscreen, r/*rectaddpt(r, screen->r.min)*/, img, nil, p2, S);
			}
	} else {
		for (x = i = 0; x < sizex; x += (pixwidth + pixspacex), ++i)
			for (y = j = 0; y < sizey; y += (pixheight + pixspacey), ++j) {
				//XSetForeground(dpy, window_gc, XGetPixel(orig_map, tlx+i, tly+j));
				//XFillRectangle(dpy, window, window_gc, 
				//   i * (pixwidth + pixspacex),
				//    j * (pixheight + pixspacey), pixwidth, pixheight);

				r = Rect(i * (pixwidth + pixspacex),
					j * (pixheight + pixspacey), 
					i * (pixwidth + pixspacex)+pixwidth, 
					j * (pixheight + pixspacey)+pixheight);
				drawop(color, color->r, img, nil, Pt(tlx+i, tly+j), S);
				drawop(lscreen, r, color, nil, ZP, S);
			}
	}
}


void screenhack(void)
{
	init_hack();
	while (1) {
		onestep();
		draw(screen, screen->r, lscreen, nil, ZP);
		flushimage(display, 1);
		frames++;

		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(0);
				}
		}
		if (delay)
			sleep(delay);
	}
}


void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	if(lscreen != nil)
		freeimage(lscreen);

	lscreen = allocimage(display, Rect(0, 0, Dx(screen->r), Dy(screen->r)), screen->chan, 1, DCyan);
	if(lscreen == nil)
		sysfatal("can't alloc buffer img: %r\n");

	draw(screen, screen->r, display->black, nil, ZP);

	init_hack();

}

void
threadmain(int argc, char *argv[])
{
	Image *glenda;
	int fd;

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(nsec());


	if(initdraw(nil, nil, "zoom") < 0)
		sysfatal("initdraw failed: %r");

	color = allocimage(display, Rect(0, 0, 1, 1), screen->chan, 1, DCyan);
	if(color == nil)
		sysfatal("can't alloc color img: %r\n");

	img = allocimage(display, display->image->r, screen->chan, 1, DCyan);
	if(img == nil)
		sysfatal("can't alloc snapshot img: %r\n");


	if(!nrand(10)) {
		fd = open("/lib/face/48x48x4/g/glenda.1", OREAD);
		if(fd < 0)
			sysfatal("cannot open /lib/face/48x48x4/g/glenda.1: %r");

		glenda = readimage(display, fd, 0);
		if(glenda == nil)
			sysfatal("cannot load glenda's image: %r");

		close(fd);

		glenda->repl = 1;
		replclipr(glenda, 1, img->r);
		draw(img, img->r, glenda, nil, ZP);
	} else {
		draw(img, img->r, display->image, nil, ZP);
	}
	einit(Emouse);
	eresized(0);

	screenhack();
}
