/* fuzzyflakes, Copyright (c) 2004
 *  Barry Dmytro <badcherry@mailc.net>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/*
 * Ported to Plan 9 by mirtchovski@gmail.com, 06/04
 */

#include "plan9.h"
#include "smooth.h"

int resized;

typedef struct _flake_var
{
	double Ticks;
	double XPos, YPos;
	double TrueX;
	double XOffset;
	double Angle;
} FlakeVariable;

static struct _flake
{ /* Struct containing the atrributes to our flakes */
	int Arms;
	int Thickness;
	int BorderThickness;
	int Radius;
	unsigned long BordColor;
	unsigned long ForeColor;
	unsigned long BackColor;
	int Layers;
	int Density;
	int Delay;
	int FallingSpeed;
	FlakeVariable *** Flakes; /* a dynamic array containing positions of all the flakes */
	struct _dbevar
	{
		Bool dbuf;
		Image *b,*ba,*bb;
	} DB;
} Flake;

void 
freeflakes(void)
{
	int i, j;

	for(i=0;i<=Flake.Layers;i++) {
		for(j=0;j<Flake.Density;j++) {
			free(Flake.Flakes[i][j]);
		}
		free(Flake.Flakes[i]);
	}
	/* this is a leak, but it breaks something in p9 :( */
	//	free(Flake.Flakes);
}
static void
InitFuzzyFlakes (void)
{
	int i,j;


	if(Flake.Flakes != nil)
		freeflakes();

	Flake.Arms		= 5;
	Flake.Thickness	= 10;
	Flake.BorderThickness	= 2;
	Flake.Radius		= 20;
	Flake.BordColor	= nrand(256);
	Flake.ForeColor	= nrand(256);
	Flake.BackColor	= nrand(256);
	Flake.Density		= 5;
	Flake.Layers		= 3;
	Flake.FallingSpeed	= 10;
	Flake.Delay		= 0;

	Flake.Density = Dx(screen->r)/200*Flake.Density;
	Flake.Flakes = malloc(sizeof(FlakeVariable**)*Flake.Layers);
	for(i=0;i<=Flake.Layers;i++)
	{
		Flake.Flakes[i] = malloc(sizeof(FlakeVariable*)*Flake.Density);
		for(j=0;j<Flake.Density;j++)
		{
			Flake.Flakes[i][j] = malloc(sizeof(FlakeVariable));
			Flake.Flakes[i][j]->XPos = nrand(Dx(screen->r));
			Flake.Flakes[i][j]->YPos = nrand(Dy(screen->r));
			Flake.Flakes[i][j]->Angle = nrand(360*(PI/180));
			Flake.Flakes[i][j]->Ticks = nrand(360);
			Flake.Flakes[i][j]->XOffset = nrand(Dy(screen->r));
		}
	}
}

static void
FuzzyFlakesMove(void)
{
	int i,j;
	for(i=1;i<=Flake.Layers;i++)
	{
		for(j=0;j<Flake.Density;j++)
		{
			FlakeVariable * FlakeVar;
			FlakeVar = Flake.Flakes[i-1][j];
			FlakeVar->Ticks++;
			FlakeVar->YPos = FlakeVar->YPos + ((double)Flake.FallingSpeed)/10/i;
			FlakeVar->TrueX = (sin(FlakeVar->XOffset + FlakeVar->Ticks*(PI/180)*((double)Flake.FallingSpeed/10)))*10 + FlakeVar->XPos;
			FlakeVar->Angle = FlakeVar->Angle + 0.005*((double)Flake.FallingSpeed/10);
			if(FlakeVar->YPos - Flake.Radius > Dy(screen->r))
			{
				FlakeVar->Ticks = 0;
				FlakeVar->YPos = 0 - Flake.Radius;
			}
		}
	}
}

static void
FuzzyFlakesDrawFlake(int XPos, int YPos, double AngleOffset, int Layer)
{
	Point p1, p2;
	int i;
	double x,y,Angle,Radius;

	/* calculate the shrink factor depending on which layer we are drawing atm */
	Radius = (double)(Flake.Radius - Layer * 5);

	/* draw the flake one arm at a time */
	for(i=1;i<=Flake.Arms;i++)
	{
		int Diameter;
		Diameter = (Flake.BorderThickness*2 + Flake.Thickness)/Layer;
		/* compute the angle of this arm of the flake */
		Angle = ((2*PI)/Flake.Arms)*i + AngleOffset;
		/* calculate the x and y dispositions for this arm */
		y=(int)(sin(Angle)*Radius);
		x=(int)(cos(Angle)*Radius);
		/* draw the base for the arm */
		//		Flake.GCValues.line_width = Diameter;
		//		XFreeGC(Flake.DisplayVar,Flake.GCVar);
		//		Flake.GCVar = XCreateGC (Flake.DisplayVar, Flake.DB.b, Flake.GCFlags, &Flake.GCValues);
		//		XSetForeground(Flake.DisplayVar,Flake.GCVar,Flake.BordColor);
		//		XDrawLine(Flake.DisplayVar, Flake.DB.b, Flake.GCVar, XPos,YPos,XPos+x,YPos+y);
		p1 = Pt(XPos, YPos);
		p2 = Pt(XPos+x, YPos+y);
		line(img, p1, p2, Endsquare, Endsquare, Diameter, colors[Flake.BordColor], ZP);

	}
	/* draw the flake one arm at a time */
	for(i=1;i<=Flake.Arms;i++)
	{
		/* compute the angle of this arm of the flake */
		Angle = ((2*PI)/Flake.Arms)*i + AngleOffset;
		/* calculate the x and y dispositions for this arm */
		y=(int)(sin(Angle)*Radius);
		x=(int)(cos(Angle)*Radius);
		/* draw the inside of the arm */
		//		Flake.GCValues.line_width = Flake.Thickness/Layer;
		//		XFreeGC(Flake.DisplayVar,Flake.GCVar);
		//		Flake.GCVar = XCreateGC (Flake.DisplayVar, Flake.DB.b, Flake.GCFlags, &Flake.GCValues);
		//		XSetForeground(Flake.DisplayVar,Flake.GCVar,Flake.ForeColor);
		//		XDrawLine(Flake.DisplayVar, Flake.DB.b, Flake.GCVar, XPos,YPos,XPos+x,YPos+y);

		p1 = Pt(XPos, YPos);
		p2 = Pt(XPos+x, YPos+y);
		line(img, p1, p2, Endsquare, Endsquare, Flake.Thickness/Layer, colors[Flake.ForeColor], ZP);
	}
}

static void
FuzzyFlakes (void)
{
	int i,j;

	FuzzyFlakesMove();


	for(i=Flake.Layers;i>=1;i--)
	{
		for(j=0;j<Flake.Density;j++)
		{
			FuzzyFlakesDrawFlake(
				Flake.Flakes[i-1][j]->TrueX,
				Flake.Flakes[i-1][j]->YPos,
				Flake.Flakes[i-1][j]->Angle,i);
		}
	}

}

void
screenhack (void)
{
	while (1)
	{
		if(resized) {
			InitFuzzyFlakes();
			resized = 0;
		}

		FuzzyFlakes ();
		draw(screen, screen->r, img, nil, ZP);
		frames++;
		flushimage(display, 1);
		draw(img, img->r, colors[Flake.BackColor], nil, ZP);
		if(ecanmouse()) {
			m = emouse();
			if(m.buttons&4)
				if(emenuhit(3, &m, &menu) == 0) {
					threadexitsall(nil);
				}
		}
		if (Flake.Delay) sleep (Flake.Delay);
	}
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0) {
		sysfatal("can't reattach to window");
	}

	resized = 1;

	if(img)
		freeimage(img);
	img = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DNofill);
	if(img == nil)
		threadexitsall("cannot allocate buffer image");

	smooth(colors);

}

void 
threadmain(int argc, char **argv)
{

	ARGBEGIN{
	case 'b':
		btime = atol(EARGF(usage()));
		proccreate(dofps, nil, 8192);
		break;
	}
	ARGEND;
	srand(time(0));


	if(initdraw(nil, nil, "fuzzyflakes") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse);

	eresized(0);
	screenhack();

}
